from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable, Type, Set

from ..tools.errors import MissingGizmoError



class UnknownSize(TypeError):
	def __init__(self):
		super().__init__('did you forget to provide a "default_len" in __init__?')



class MissingBuffer(MissingGizmoError):
	pass



class BudgetExceeded(StopIteration):
	pass



class EpochEnd(StopIteration):
	pass







from omniply.core.abstract import *


class ApplicationAmbiguityError(ValueError):
	def __init__(self, gizmo: str, options: Sequence[str], *, message: Optional[str] = None):
		if message is None:
			message = f'{gizmo!r} cant choose between: {options}'
		super().__init__(message)
		self.gizmo = gizmo
		self.options = options



class MissingValueError(AttributeError):
	def __init__(self, descriptor: Any, message: Optional[str] = None):
		if message is None:
			message = f'{descriptor}'
		super().__init__(message)



class MissingOwnerError(TypeError):
	def __init__(self, descriptor: Any, message: Optional[str] = None):
		if message is None:
			message = f'Cannot delete {descriptor} without providing an instance.'
		super().__init__(message)



class ReachFailedFlag(Exception):
	def __init__(self):
		super().__init__('to access this "realize()" is necessary')



class IgnoreResetFlag(Exception):
	pass



class GadgetFailedError(AbstractGadgetError):
	def __init__(self, gizmo: str, *, message: Optional[str] = None):
		if message is None:
			message = f'{gizmo!r}'
		super().__init__(message)
		self.gizmo = gizmo


	def reason(self):
		return f'{self.gizmo!r} {self.__class__.__name__}'



class AssemblyFailedError(GadgetFailedError):
	def __init__(self, gizmo: str, *failures: GadgetFailedError,
                 message: Optional[str] = None):
		if message is None:
			message = self.verbalize_failures(gizmo, failures)
		super().__init__(gizmo, message=message)
		self.gizmo = gizmo
		self.failures = failures


	def reason(self):
		return f'{self.gizmo} [{", ".join(f.reason() if isinstance(f, GadgetFailedError) else repr(f) for f in self.failures)}]'


	@staticmethod
	def verbalize_failures(gizmo: str, failures: Sequence[GadgetFailedError]):
		# path = [repr(e.gizmo) for e in failures]
		# terminal = failures[-1].reason()
		reason = f'[{", ".join(f.reason() if isinstance(f, GadgetFailedError) else repr(f) for f in failures)}]'
		return f'{gizmo!r} due to {reason}'



class MissingGizmoError(GadgetFailedError, KeyError):
	def __init__(self, gizmo: str, *, message: Optional[str] = None):
		if message is None:
			message = gizmo
		super().__init__(gizmo, message=message)


	def reason(self):
		return f'missing {self.gizmo!r}'



# class SkipToolFlag(ToolFailedError):
# 	'''raised by a tool to indicate that it should be skipped'''
# 	pass












from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable, Type, Set

from omnibelt import method_decorator, agnostic, unspecified_argument, filter_duplicates
from omnibelt.crafts import AbstractCraft, AbstractCrafty, NestableCraft, SkilledCraft, IndividualCrafty, HiddenCrafty

from ..features import Prepared
from ..structure import spaces

from .abstract import AbstractSpaced, Loggable, AbstractAssessible, AbstractKit, SingleVendor, AbstractTool, \
	AbstractChangableSpace, AbstractDynamicKit
from .errors import ToolFailedError, MissingGizmoError
from .crafts import ToolCraft, OptionalCraft, DefaultCraft, LabelCraft, SpaceCraft, InitCraft, ReplaceableCraft
from .assessments import AbstractSignature, Signatured


class DynamicKit(AbstractDynamicKit):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._added_tools = []


	def include(self, *tools: AbstractTool) -> 'AbstractDynamicKit':
		self._added_tools.extend(tools)
		return self


	def tools(self) -> Iterator['AbstractTool']:
		yield from getattr(self, '_added_tools', [])



class SpacedTool(AbstractTool, AbstractSpaced):
	def space_of(self, gizmo: str) -> spaces.Dim:
		for tool in self.vendors(gizmo):
			try:
				if isinstance(tool, AbstractSpaced):
					return tool.space_of(gizmo)
			except ToolFailedError:
				pass
		raise self._ToolFailedError(f'No tool for {gizmo} in {self}')



class SpaceKit(IndividualCrafty, AbstractChangableSpace): # processes `space` decorators
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._spaces = {}


	def clear_space(self, gizmo: str) -> bool:
		if gizmo in self._spaces:
			self._spaces[gizmo][0].clear_space(gizmo)


	def gizmos(self) -> Iterator[str]:
		yield from self._spaces.keys()


	def _process_skill(self, src: Type[AbstractCrafty], key: str, craft: AbstractCraft, skill: LabelCraft.Skill):
		super()._process_skill(src, key, craft, skill)
		if isinstance(skill, SpaceCraft.Skill):
			self._spaces.setdefault(skill.label, []).append(skill)


	def space_of(self, gizmo: str):
		if gizmo in self._spaces:
			return self._spaces[gizmo][0].space_of(gizmo)
		# return self._tools[gizmo].space_of(gizmo)
		return super().space_of(gizmo)


	def _default_space_of(self, gizmo: str):
		raise NotImplementedError # TODO

	
	# def gizmos(self) -> Iterator[str]:
	# 	yield from filter_duplicates(super().gizmos(), self._spaces.keys())


	def change_space_of(self, gizmo: str, space: spaces.Dim):
		if gizmo in self._spaces:
			self._spaces[gizmo][0].change_space_of(gizmo, space)
		else:
			super().change_space_of(gizmo, space)

	
	# def gizmos(self) -> Iterator[str]:
	# 	yield from filter_duplicates(super().gizmos(), self._spaces.keys())



class ValidatedCrafty(IndividualCrafty):
	@staticmethod
	def validate_label(label):
		return label



class RelabeledCrafty(IndividualCrafty):
	'''Relabel inherited crafts'''
	_inherited_tool_relabels = None
	def __init_subclass__(cls, replace=None, **kwargs): # {old_label: new_label}
		if replace is None:
			replace = {}
		super().__init_subclass__(**kwargs)
		# past = {}
		# for parent in cls.__bases__:
		# 	if issubclass(parent, RelabeledKit) and parent._inherited_tool_relabels is not None:
		# 		past.update(parent._inherited_tool_relabels)
		# replace.update(past)
		cls._inherited_tool_relabels = replace


	@agnostic
	def _emit_all_craft_items(self, *, remaining: Iterator[Type['InheritableCrafty']] = None,
	                          start : Type['InheritableCrafty'] = None, owner : Type['InheritableCrafty'] = None,
	                          **kwargs) -> Iterator[Tuple[Type[AbstractCrafty], str, AbstractCraft]]: # N-O
		live = start is None # make sure replacements only happen once
		cls = self if isinstance(self, type) else type(self)
		if start is None:
			start = cls

		for loc, key, craft in super()._emit_all_craft_items(remaining=remaining, start=start, owner=owner, **kwargs):
			if loc is start:
				yield loc, key, craft
			elif live and len(start._inherited_tool_relabels) and isinstance(craft, ReplaceableCraft):
				# if issubclass(loc, RelabeledKit):
				# 	craft = craft.replace(loc._inherited_tool_relabels)
				fix = craft.replace(start._inherited_tool_relabels)
				yield loc, key, fix
			else:
				yield loc, key, craft



class ElasticCrafty(ValidatedCrafty):
	'''Relabel instance crafts'''
	_application = None

	def __init__(self, *args, application=None, **kwargs):
		if application is None:
			application = {}
		super().__init__(*args, **kwargs)
		self._application = application

	@agnostic
	def validate_label(self, label):
		if self._application is None:
			return label
		return self._application.get(label, label)



class CraftyKit(SpaceKit, SpacedTool, AbstractKit):
	class _SkillTool(AbstractTool): # collects all skills (of the whole mro) of one gizmo
		def __init__(self, label: str, **kwargs):
			super().__init__(**kwargs)
			self._label = label
			self._standard = []
			self._optionals = []
			self._defaults = []


		def add(self, skill: ToolCraft.Skill):
			if isinstance(skill, OptionalCraft.Skill):
				self._optionals.append(skill)
			elif isinstance(skill, DefaultCraft.Skill):
				self._defaults.append(skill)
			else:
				self._standard.append(skill)


		def tool_history(self):
			yield from self._standard
			yield from self._optionals
			yield from self._defaults


		def tool(self):
			return next(self.tool_history())


		def tools(self): # by default, only uses the newest tool in the class hierarchy
			yield self.tool()


	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._tool_skills = {}


	def _process_skill(self, src: Type[AbstractCrafty], key: str, craft: AbstractCraft, skill: LabelCraft.Skill):
		super()._process_skill(src, key, craft, skill)
		if isinstance(skill, ToolCraft.Skill):
			if skill.label not in self._tool_skills:
				self._tool_skills[skill.label] = self._SkillTool(skill.label)
			self._tool_skills[skill.label].add(skill)


	def has_gizmo(self, gizmo: str) -> bool:
		return gizmo in self._tool_skills


	def vendors(self, gizmo: str):
		# gizmo = self.validate_label(gizmo)
		if gizmo not in self._tool_skills:
			raise MissingGizmoError(gizmo)
		yield from self._tool_skills[gizmo].tools()


	def tools(self) -> Iterator['AbstractTool']:
		for skill in self._tool_skills.values():
			yield from skill.tools()


	def gizmos(self) -> Iterator[str]:
		yield from filter_duplicates(self._spaces.keys(), super(SpaceKit, self).gizmos())



class MaterialedCrafty(CraftyKit, Prepared): # allows materials to be initialized when prepared
	def _prepare(self, *args, **kwargs):
		super()._prepare(*args, **kwargs)

		materials = {}
		for gizmo, tool in self._tool_skills.items():
			if isinstance(tool, self._SkillTool):
				skill = tool.tool()
				if isinstance(skill, InitCraft.Skill):
					materials[gizmo] = skill.init(self)
		self._tool_skills.update(materials)



class SignaturedCrafty(CraftyKit, Signatured):
	@agnostic
	def signatures(self, owner = None) -> Iterator['AbstractSignature']:
		outputs = set()
		if isinstance(self, type):
			self: Type['SignaturedCrafty']
			for loc, key, craft in self._emit_all_craft_items():
				if isinstance(craft, Signatured):
					for signature in craft.signatures(self):
						if signature.output not in outputs:
							outputs.add(signature.output)
							yield signature
		else:
			for tool in self.tools():
				if isinstance(tool, Signatured):
					yield from tool.signatures(self)



class AssessibleCrafty(CraftyKit, AbstractAssessible):
	@classmethod
	def assess_static(cls, assessment):
		super().assess_static(assessment)
		for owner, key, craft in cls._emit_all_craft_items():
			if isinstance(craft, AbstractAssessible):
				assessment.add_edge(cls, craft, name=key, loc=owner)
				assessment.expand(craft)


	def assess(self, assessment):
		super().assess(assessment)
		for tool in self.tools():
			if isinstance(tool, AbstractAssessible):
				assessment.add_edge(self, tool)
				assessment.expand(tool)

















# import torch
# from omnibelt import unspecified_argument, duplicate_instance, InitWall



class WrappedTensor():#torch.Tensor):
	@staticmethod
	def __new__(cls, src, *args, **kwargs):
		return super().__new__(cls, src)#, *args, **kwargs)


	def __init__(self, src=None, *args, **kwargs):
		# super().__init__(src, *args, **kwargs)
		pass


	def copy(self):
		copy = self.__class__.__new__(self.__class__, [])
		copy.__dict__.update(self.__dict__)
		copy.data = self.data
		return copy


	def clone(self, *args, **kwargs):
		new = self.copy()
		new.data = super().clone(*args, **kwargs)
		return new


	def to(self, *args, **kwargs):
		new = self.copy()
		new.data = super().to(*args, **kwargs)
		return new



class SpaceTensor(WrappedTensor):
	def __init__(self, src, space=None, **kwargs):
		super().__init__(src, **kwargs)
		self.space = space








import numpy as np


def angle_diff(angle1, angle2, period=2*np.pi):
	a = angle1 - angle2
	return (a + period/2) % period - period/2



def round_sigfigs(x, sigfigs=3):
	mag = x.abs().log10().floor()
	mag[mag.isinf()] = 0
	reg = 10 ** (sigfigs - mag - 1)
	return x.mul(reg).round().div(reg)



def sigfig_noise(x, n, sigfigs=3):
	mag = x.abs().log10().floor()
	mag[mag.isinf()] = 0
	reg = 10 ** (sigfigs - mag - 1)
	return x.mul(reg).add(n).div(reg)



def mixing_score(mat, dim1=1, dim2=2, eps=1e-10): # TODO: what is this?
	N1, N2 = mat.size(dim1), mat.size(dim2)
	mat = mat.abs()
	return 1 - mat.div(mat.add(eps).max(dim1, keepdim=True)[0]).sum(dim1, keepdim=True).sub(1)\
		.mean(dim2, keepdim=True).div(N1-1)





from omnibelt import unspecified_argument, InitWall
# import torch
# from torch.nn import functional as F
# import torch.distributions as distrib
# from torch.distributions.utils import lazy_property
# from torch.distributions import constraints

from ..util.tensors import WrappedTensor
from ..features import Device
from ..persistent import Fingerprinted

from .operations import Sampler


# TODO: make exportable
# TODO: add warnings for non-seeded functions like rsample()
class Distribution(Sampler, Fingerprinted, ):#distrib.Distribution): # TODO: saving and loading distributions
	def __init__(self, params=None, apply_constraints=False, constraints=None, soft_constraints=False, **kwargs):
		if params is None:
			params = {key: kwargs[key] for key in self.arg_constraints if key in kwargs}
		if constraints is None:
			constraints = self.arg_constraints
		kwargs.update(self._constrain_params(params, constraints, soft_constraints=soft_constraints)
		              if apply_constraints else params)
		super().__init__(**kwargs)
		self._param_keys = set(key for key in params if params[key] is not None)


	def _constrain_params(self, params, constraints, **kwargs):
		return {key: self.constrain_value(constraints.get(key), val, **kwargs) for key, val in params.items()}


	# def sample(self, sample_shape=torch.Size()):
	# 	if isinstance(sample_shape, int):
	# 		sample_shape = sample_shape,
	# 	return super().sample(sample_shape)


	def best(self):
		'''
		Get the best sample (by default, the mean, but possibly also the mode)
		:return:
		'''
		return self.mean


	def iterate_parameters(self):
		for key in self._param_keys:
			val = getattr(self, key)
			if val is not None:
				yield (key, val)


	def _fingerprint_data(self):
		params = {key: self.fingerprint_obj(getattr(self, key)) for key in self._param_keys if hasattr(self, key)}
		return {'params': params, **super()._fingerprint_data()}


	def copy(self, **kwargs):
		for param in self._param_keys:
			if param not in kwargs:
				kwargs[param] = getattr(self, param)
		return self.__class__(**kwargs)


	def to(self, device):
		params = {}
		for key in self._param_keys:
			param = getattr(self, key)
			if isinstance(param, (torch.Tensor, Device)):
				param = param.to(device)
			params[key] = param
		return self.copy(**params)


	@staticmethod
	def constrain_value(constraint, value, eps=1e-12, soft_constraints=False):
		if value is None:
			return value

		if constraint == constraints.simplex:
			return F.normalize(value, p=1, dim=1, eps=eps).abs()

		if isinstance(constraint,
		              (constraints.greater_than, constraints.greater_than_eq)) and constraint.lower_bound == 0:
			return F.softplus(value) if soft_constraints else value.exp()

		if isinstance(constraint, constraints.interval):
			mn, mx = constraint.lower_bound, constraint.upper_bound
			# assert mn == 0 and mx == 1, f'{constraint}: {mn} {mx}'
			value = value.sigmoid()
			if mn != 0 or mx != 1:
				value = value.mul(mx - mn) + mn
			return value

		if constraint == constraints.positive_definite:
			raise NotImplementedError  # TODO
			if len(real.shape) == 1:
				N = int(np.sqrt(len(real)))
				real = value.reshape(N, N) + ep
				s * torch.eye(N, device=real.device)
			return real.t() @ real

		return value



class NormalDistribution(Distribution, ):#distrib.Normal):
	def __init__(self, loc, scale, **kwargs):
		super().__init__(loc=loc, scale=scale, **kwargs)


	def _sample(self, shape):
		shape = self._extended_shape(shape)
		noise = torch.randn(*shape, generator=self.rng)
		return noise.mul(self.scale) + self.loc



class CategoricalDistribution(Distribution, ):#distrib.Categorical):
	def __init__(self, probs=None, logits=None, **kwargs):
		super().__init__(probs=probs, logits=logits, **kwargs)

	def best(self):
		return self._param.argmax(-1)


	def _sample(self, shape):
		probs_2d = self.probs.reshape(-1, self._num_events)
		samples_2d = torch.multinomial(probs_2d, shape.numel(), True, generator=self.rng).T
		return samples_2d.reshape(self._extended_shape(shape))



class DistributionTensor(Fingerprinted, WrappedTensor):

	_base_distribution = None
	@classmethod
	def _create_distribution(cls, **kwargs):
		return cls._base_distribution(**kwargs)


	@staticmethod
	def __new__(cls, dis=None, value=None, use_best=True, **kwargs):

		if dis is None:
			dis = cls._create_distribution(**kwargs)

		if value is None:
			value = dis.best() if use_best else dis.generate()

		obj = super().__new__(cls, value.float())
		obj._distribution = dis
		return obj


	def _fingerprint_data(self):
		return {'distribution': self.fingerprint_obj(self._distribution), **super()._fingerprint_data()}


	# def __repr__(self):
	# 	return repr(self.distribution)

	def copy(self):
		params = dict(self.distribution.iterate_parameters())
		copy = self.__class__.__new__(self.__class__, value=self, **params)
		copy.__dict__.update(self.__dict__)
		copy.data = self.data
		return copy


	@property
	def distribution(self):
		return self._distribution


	def to(self, *args, **kwargs):
		try:
			new = super().to(*args, **kwargs)
		except RuntimeError:
			print('Cannot change the dtype of distributions (only devices)') # TODO
			return self
		else:
			new._distribution = self.distribution.to(*args, **kwargs)
			return new


	# def __getattr__(self, item):
	# 	value = getattr(self.distribution, item, unspecified_argument)
	# 	if value is unspecified_argument:
	# 		raise AttributeError(item)
	# 	return value



class Normal(DistributionTensor):
	_base_distribution = NormalDistribution

	@staticmethod
	def __new__(cls, loc, scale, **kwargs):
		return super().__new__(cls, loc=loc, scale=scale, **kwargs)



class Categorical(DistributionTensor):
	_base_distribution = CategoricalDistribution

	@staticmethod
	def __new__(cls, probs=None, logits=None, apply_constraints=False, soft_constraints=False,
	            gen=unspecified_argument, seed=unspecified_argument, **kwargs):
		return super().__new__(cls, probs=probs, logits=logits, **kwargs)






from omnibelt import agnostic, unspecified_argument
import os
import random
import inspect
import numpy as np
# import torch


class RNGManager:
	_fallback_rng_seed = 16283393149723337453
	def __init__(self, *args, master_seed=None, **kwargs):
		if master_seed is None:
			master_seed = self.gen_random_seed()
		super().__init__(*args, **kwargs)
		self._meta_rng = None
		self._forced_rngs = []
		self._default_rngs = []
		self.master_seed = master_seed
		self.set_global_seed(self.master_seed)

	def push_forced(self, seed=None, *, rng=None):
		if rng is None:
			rng = self.create_rng(seed=seed)
		self._forced_rngs.append(rng)
	def pop_forced(self):
		return self._forced_rngs.pop()

	def push_default(self, seed=None, *, rng=None):
		if rng is None:
			rng = self.create_rng(seed=seed)
		self._default_rngs.append(rng)
	def pop_default(self):
		return self._default_rngs.pop()

	@property
	def master_seed(self):
		return self._master_seed
	@master_seed.setter
	def master_seed(self, seed):
		self._master_seed = seed
		self._meta_rng = self.create_rng(seed=seed)

	def create_personal_rng(self, obj, seed=unspecified_argument):
		if seed is unspecified_argument:
			seed = getattr(obj, '_seed', None)
		rng = self.create_rng(seed=seed, base_gen=self._meta_rng)
		obj._rng = rng
		return rng

	def get_default_rng(self):
		if len(self._default_rngs):
			return self._default_rngs[0]

	def get_personal_rng(self, obj=None):
		if len(self._forced_rngs):
			return self._forced_rngs[-1]

		personal = getattr(obj, '_rng', None)
		if personal is None:
			return self.create_personal_rng(obj)
		return personal

	def replace_personal_rng(self, obj, *, rng=None, seed=unspecified_argument):
		if rng is None and seed is not unspecified_argument:
			rng = self.create_rng(seed=seed)
		obj._rng = rng
		return rng

	def clear_personal_rng(self, obj):
		self.replace_personal_rng(obj, rng=None)

	@agnostic
	def create_rng(self, seed=None, base_gen=None):
		if seed is None:
			seed = self.gen_random_seed(base_gen)
		return self._create_rng(seed)

	@agnostic
	def gen_deterministic_seed(self, base_seed):
		return self.gen_random_seed(self.create_rng(base_seed))

	@agnostic
	def gen_random_seed(self, gen=None) -> int:
		raise NotImplementedError

	def _create_rng(self, seed):
		raise NotImplementedError

	def set_global_seed(self, seed):
		os.environ["PYTHONHASHSEED"] = str(seed)

		np.random.seed(seed % (2**32))
		random.seed(seed)

	def force_rng(self, seed=unspecified_argument, *, rng=None):
		if seed is not unspecified_argument or rng is not None:
			return self.SeedContext(self.push_forced, self.pop_forced, self, seed=seed, rng=rng)

	def default_rng(self, seed=unspecified_argument, *, rng=None):
		if seed is not unspecified_argument or rng is not None:
			return self.SeedContext(self.push_default, self.pop_default, self, seed=seed, rng=rng)

	class SeedContext:
		def __init__(self, push, pop, src=None, seed=None, rng=None):
			self.push = push
			self.pop = pop
			self.src = src
			self.seed = seed
			self.rng = rng

		def __enter__(self):
			if self.rng is None:
				self.rng = self.src.create_rng(seed=self.seed)
			self.push(rng=self.rng)
			return self.rng

		def __exit__(self, exc_type, exc_val, exc_tb):
			self.pop()


import numpy as np

class PytorchManager(RNGManager):
	@agnostic
	def gen_random_seed(self, gen=None):
		if gen is None:
			gen = np.random
		return gen.randint(0, 2**30)#.item()
		# return torch.randint(-2 ** 63, 2 ** 63 - 1, size=(), generator=gen).item()

	@agnostic
	def _create_rng(self, seed):
		# gen = torch.Generator()
		# gen.manual_seed(seed)
		gen = np.random.RandomState(seed)
		return gen

	@agnostic
	def set_global_seed(self, seed):
		super().set_global_seed(seed)
		np.random.seed(seed)
		# torch.manual_seed(seed)
		# torch.cuda.manual_seed(seed)
		# torch.backends.cudnn.deterministic = True
		# torch.backends.cudnn.benchmark = False


# class NumpyManager(RNGManager):
# 	pass

# class TensorflowManager(RNGManager):
# 	pass

	# tf.random.set_seed(seed)
	# tf.experimental.numpy.random.seed(seed)
	# tf.set_random_seed(seed)
	# os.environ['TF_CUDNN_DETERMINISTIC'] = '1'
	# os.environ['TF_DETERMINISTIC_OPS'] = '1'


default_rng_manager = PytorchManager()


def force_rng(seed=unspecified_argument, *, rng=None):
	return default_rng_manager.force_rng(seed=seed, rng=rng)


def default_rng(seed=unspecified_argument, *, rng=None):
	return default_rng_manager.default_rng(seed=seed, rng=rng)


def gen_deterministic_seed(base_seed):
	return default_rng_manager.gen_deterministic_seed(base_seed)


def gen_random_seed(base_gen=None):
	return default_rng_manager.gen_random_seed(base_gen)


def create_rng(seed=None, base_gen=None):
	return default_rng_manager.create_rng(seed=seed, base_gen=base_gen)


def set_global_seed(seed):
	return default_rng_manager.set_global_seed(seed)



class RNG_Base: # descriptor
	Manager = PytorchManager

	manager = default_rng_manager

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		if self.manager is None:
			self.__class__.manager = self.Manager()

	def reset(self, obj): # TODO
		self.manager.clear_personal_rng(obj)
		return self.manager.get_personal_rng(obj)

	def force_rng(self, seed=unspecified_argument, *, rng=None):
		return self.manager.force_rng(seed=seed, rng=rng)

	def default_rng(self, seed=unspecified_argument, *, rng=None):
		return self.manager.default_rng(seed=seed, rng=rng)

	def get_default_rng(self):
		return self.manager.get_default_rng()

	def gen_deterministic_seed(self, base_seed):
		return self.manager.gen_deterministic_seed(base_seed)

	def gen_random_seed(self, base_gen=None):
		return self.manager.gen_random_seed(base_gen)



class RNG(RNG_Base):
	def __get__(self, obj, owner=None):
		return self.manager.get_personal_rng(obj)

	def __set__(self, obj, value):
		self.manager.replace_personal_rng(obj, rng=value)



class RNG_Link(RNG_Base):
	def __get__(self, obj, owner=None):
		return self.manager.get_default_rng()

	def __set__(self, obj, value):
		pass



class Seedable: # has no rng of its own, but respects forced rngs
	rng = RNG_Link()
	_seed = None

	def __init__(self, *args, rng=unspecified_argument, seed=unspecified_argument, **kwargs):
		super().__init__(*args, **kwargs)
		if seed is not unspecified_argument:
			self._seed = seed
		if rng is not unspecified_argument:
			self.rng = rng

	def default_rng(self, seed=unspecified_argument, *, rng=None):
		if rng is None:
			rng = self.rng
		return self._get_rng().default_rng(seed=seed, rng=rng)

	def force_rng(self, seed=unspecified_argument, *, rng=None):
		if rng is None:
			rng = self.rng
		return self._get_rng().force_rng(seed=seed, rng=rng)

	@classmethod
	def _get_rng(cls):
		return inspect.getattr_static(cls, 'rng', None)

	def reset_rng(self, seed=unspecified_argument):
		if seed is not unspecified_argument:
			self._seed = seed
		return self._get_rng().reset(self)

	@property
	def seed(self):
		return self._seed



class Seeded(Seedable): # uses its own RNG unless one is forced
	rng = RNG() # access only when needed (rather than passing as argument)




# def set_global_seed(seed=None):
# 	if seed is None:
# 		seed = Seeded.gen_random_seed()
# 	random.seed(seed)
# 	np.random.seed(seed)
# 	torch.manual_seed(seed)
# 	if torch.cuda.is_available():
# 		torch.cuda.manual_seed(seed)
# 	return seed


# def set_seed(seed: int = 42) -> None:
#     np.random.seed(seed)
#     random.seed(seed)
#     torch.manual_seed(seed)
#     torch.cuda.manual_seed(seed)
#     # When running on the CuDNN backend, two further options must be set
#     torch.backends.cudnn.deterministic = True
#     torch.backends.cudnn.benchmark = False
#     # Set a fixed value for the hash seed
#     os.environ["PYTHONHASHSEED"] = str(seed)
#     print(f"Random seed set as {seed}")

# def set_seed(seed: int = 42) -> None:
#   random.seed(seed)
#   np.random.seed(seed)
#   tf.random.set_seed(seed)
#   tf.experimental.numpy.random.seed(seed)
#   tf.set_random_seed(seed)
#   # When running on the CuDNN backend, two further options must be set
#   os.environ['TF_CUDNN_DETERMINISTIC'] = '1'
#   os.environ['TF_DETERMINISTIC_OPS'] = '1'
#   # Set a fixed value for the hash seed
#   os.environ["PYTHONHASHSEED"] = str(seed)
#   print(f"Random seed set as {seed}")














from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Generator, Type, Iterable, \
	Iterator, NamedTuple, ContextManager
from omnibelt import agnostic, unspecified_argument

from .errors import MissingBuilderError, NoProductFound, InvalidProductError


class AbstractHyperparameter:
	def __get__(self, instance, owner):
		raise NotImplementedError


	def validate(self, instance: 'AbstractParameterized') -> Any:
		raise NotImplementedError



class AbstractParameterized:
	def fill_hparams(self, fn, args=None, kwargs=None, **finder_kwargs) -> Dict[str, Any]:
		raise NotImplementedError


	def _extract_hparams(self, kwargs):
		raise NotImplementedError


	def _existing_hparams(self, *, hidden=False):
		for key, param in self.named_hyperparameters(hidden=hidden):
			if param.is_missing(self):
				yield key, getattr(self, key)


	def has_hparam(self, key):
		raise NotImplementedError


	def get_hparam(self, key, default=unspecified_argument):
		raise NotImplementedError


	@classmethod
	def hyperparameters(cls, *, hidden=False):
		for key, val in cls.named_hyperparameters(hidden=hidden):
			yield val


	@classmethod
	def named_hyperparameters(cls, *, hidden=False):
		raise NotImplementedError


	@classmethod
	def hyperparameter_names(cls, *, hidden=False):
		for key, val in cls.named_hyperparameters(hidden=hidden):
			yield key


	@classmethod
	def inherit_hparams(cls, *names):
		raise NotImplementedError


	# def __setattr__(self, key, value):
	# 	if isinstance(value, AbstractHyperparameter):
	# 		raise ValueError('Hyperparameters must be set to the class (not an instance)')
	# 	super().__setattr__(key, value)



class AbstractBuilder(AbstractParameterized):
	@staticmethod
	def validate(product):
		return product


	@staticmethod
	def product(*args, **kwargs) -> Type:
		raise NotImplementedError


	@staticmethod
	def build(*args, **kwargs):
		raise NotImplementedError



class AbstractArgumentBuilder(AbstractBuilder):
	def build(self, *args, **kwargs):
		product = self.product(*args, **kwargs)
		kwargs = self._build_kwargs(product, *args, **kwargs)
		args = kwargs.pop(None) if None in kwargs else ()
		return product(*args, **kwargs)


	def _build_kwargs(self, product, *args, **kwargs):
		return kwargs.copy()



class AbstractMultiBuilder(AbstractArgumentBuilder, AbstractParameterized):
	ident = None
	_NoProductFound = NoProductFound


	@classmethod
	def product_names(cls):
		for name, product in cls.named_products():
			yield name


	@classmethod
	def products(cls):
		for name, product in cls.named_products():
			yield product


	@classmethod
	def named_products(cls):
		raise NotImplementedError


	def _build_kwargs(self, product, ident: Optional[str] = unspecified_argument, **kwargs):
		return super()._build_kwargs(product, **kwargs)


	_InvalidProductError = InvalidProductError
	def validate(self, product):
		if isinstance(product, str):
			# raise self._InvalidProductError(f'Expected product, got string: {product!r}')
			return self.build(product)
		return product



class AbstractRegistryBuilder(AbstractMultiBuilder):
	@classmethod
	def register_product(cls, name, product, *, is_default=False, **kwargs):
		raise NotImplementedError



class AbstractModular(AbstractParameterized):
	@classmethod
	def named_submodules(cls, *, hidden=False) -> Iterator[Tuple[str, 'AbstractSubmodule']]:
		for name, param in cls.named_hyperparameters(hidden=hidden):
			if isinstance(param, AbstractSubmodule):
				yield name, param


	@classmethod
	def submodules(cls, *, hidden=False) -> Iterator['AbstractSubmodule']:
		for name, param in cls.named_submodules(hidden=hidden):
			yield param


	@classmethod
	def submodule_names(cls, *, hidden=False) -> Iterator[str]:
		for name, param in cls.named_submodules(hidden=hidden):
			yield name



class AbstractSubmodule(AbstractHyperparameter):

	def init(self, instance):
		raise NotImplementedError


	# def get_builder(self, *args, **kwargs) -> Optional[AbstractBuilder]:
	# 	raise NotImplementedError


	# def validate(self, product, *, spec=None):
	# 	builder = self.get_builder() if spec is None else self.get_builder(blueprint=spec)
	# 	if builder is None:
	# 		return super().validate(product)
	# 	try:
	# 		return builder.validate(product)
	# 	except InvalidProductError:
	# 		return self.build_with(product) if spec is None else self.build_with_spec(product, spec=spec)


	# def build_with_spec(self, owner, spec=None): # TODO: --> architect
	# 	if spec is None:
	# 		spec = owner.as_spec()
	#
	# 	builder = self.get_builder()
	# 	if builder is None:
	# 		raise self._MissingBuilderError(f'No builder for {self}')
	# 	return builder.build(*args, **kwargs)



# class AbstractSpec(Iterable):
# 	def get(self, name, default=unspecified_argument) -> 'AbstractSpec':
# 		raise NotImplementedError
#
# 	@property
# 	def base(self):
# 		raise NotImplementedError
#
# 	@property
# 	def name(self):
# 		raise NotImplementedError
#
# 	@property
# 	def info(self):
# 		raise NotImplementedError
#
# 	@property
# 	def is_default(self):
# 		raise NotImplementedError






# import torch

from omnibelt import agnostic
from ..features import Seedable


class Extractor:
	@agnostic
	def extract(self, observation):
		raise NotImplementedError



class Encoder(Extractor):
	@agnostic
	def extract(self, observation):
		return self.encode(observation)


	@agnostic
	def encode(self, observation):
		raise NotImplementedError



class Decoder:
	@agnostic
	def decode(self, latent):
		raise NotImplementedError



class Generator:
	@agnostic
	def generate(self, N):
		raise NotImplementedError



class Discriminator:
	@agnostic
	def judge(self, observation):
		raise NotImplementedError



class Augmentation:
	@agnostic
	def augment(self, observation):
		raise NotImplementedError


# indicator - even more general, just a function that takes an observation and returns a (scalar) value


class Criterion: # relates to difference between two things
	@agnostic
	def compare(self, observation1, observation2):
		raise NotImplementedError



class Metric(Criterion): # obeys triangle inequality
	@agnostic
	def distance(self, observation1, observation2):
		raise NotImplementedError


	@agnostic
	def compare(self, observation1, observation2):
		return self.distance(observation1, observation2)



class PathCriterion(Criterion):
	@agnostic
	def compare(self, observation1, observation2):
		return self.compare_path(observation1, observation2)


	@agnostic
	def compare_path(self, path1, path2):
		raise NotImplementedError



class Interpolator: # returns N steps to get from start to finish ("evenly spaces", by default)
	@staticmethod
	def interpolate(start, end, N):
		raise NotImplementedError



class Estimator:
	@agnostic
	def predict(self, observation):
		raise NotImplementedError



class Invertible:
	@agnostic
	def forward(self, observation):
		raise NotImplementedError


	@agnostic
	def inverse(self, observation):
		raise NotImplementedError



class Compressor:
	@staticmethod
	def compress(observation):
		raise NotImplementedError


	@staticmethod
	def decompress(data):
		raise NotImplementedError



class Quantizer:
	@staticmethod
	def quantize(observation): # generally "removes" noise
		raise NotImplementedError


	@staticmethod
	def dequantize(observation): # generally adds noise
		raise NotImplementedError



class Sampler(Seedable):
	def sample(self, *shape, **kwargs):
		return self._sample(torch.Size(shape), **kwargs)


	def _sample(self, shape, **kwargs):
		raise NotImplementedError



from typing import Sequence as _Sequence
from omnibelt import Packable

# import math

import numpy as np
# import torch
# from torch.nn import functional as F

from ..util import angle_diff
from ..persistent import Fingerprinted

from .operations import Sampler
from .metrics import Measure, Lp, L2, L1


# TODO: include dtypes
# TODO: equality testing


class Dim(Packable, Measure, Sampler, Fingerprinted):
	def __init__(self, min=None, max=None, shape=(), dtype=None, **kwargs):
		super().__init__(**kwargs)
		self.shape = shape
		self.dtype = dtype
		self.min = min
		self.max = max


	def __str__(self):
		terms = ', '.join(map(str,self.shape))
		return f'{self.__class__.__name__}({terms})'


	def __repr__(self):
		return str(self)


	def compress(self, vals):
		return vals


	def expand(self, vals):
		return vals


	class InvalidValue(ValueError):
		pass


	def validate(self, value):
		if not isinstance(value, (int, float)):
			raise NotImplementedError
		if self.min is not None and value < self.min:
			raise self.InvalidValue(value)
		if self.max is not None and value > self.max:
			raise self.InvalidValue(value)

	@property
	def width(self): # for (default) neural representation
		return self.dof

	@property
	def dof(self): # minimal representation
		return len(self)


	@property
	def shape(self): # native representation
		return self._shape
	@shape.setter
	def shape(self, shape):
		if isinstance(shape, int):
			shape = (shape,)
		self._shape = shape# torch.Size(shape)


	@property
	def expanded_shape(self):
		return self.shape


	def __len__(self):  # compressed shape
		return self.shape.numel()


	def _fingerprint_data(self):
		return {
			'min': None if self.min is None else self.min.float().mean().item(),
			'max': None if self.max is None else self.max.float().mean().item(),
			'dtype': str(self.dtype),
			'shape': str(self.shape),
			**super()._fingerprint_data()
		}
	

	def sample(self, *shape, **kwargs):
		samples = super().sample(*shape, **kwargs)
		if self.dtype is not None:
			samples = samples.to(self.dtype)
		return samples


	def difference(self, x, y, standardize=None):  # x-y
		if standardize:
			x, y = self.standardize(x), self.standardize(y)
		return super().difference(x, y)


	def distance(self, x, y, standardize=None):
		if standardize:
			x, y = self.standardize(x), self.standardize(y)
		return super().distance(x, y)


	def measure(self, x, y, standardize=None):
		if standardize:
			x, y = self.standardize(x), self.standardize(y)
		return super().measure(x, y)


	def transform(self, vals, spec): # spec is current space of vals, self is the new
		return self.unstandardize(spec.standardize(vals))


	def standardize(self, vals):
		raise NotImplementedError


	def unstandardize(self, vals):
		raise NotImplementedError


	@property
	def min(self):
		return self._min
	@min.setter
	def min(self, val):
		self._min = val if val is None else val#torch.as_tensor(val)


	@property
	def max(self):
		return self._max
	@max.setter
	def max(self, val):
		self._max = val if val is None else val#torch.as_tensor(val)


	@property
	def range(self):
		return self.max - self.min



class Range(Dim):
	def __init__(self, min=None, max=None, shape=(), dtype=None, **kwargs):
		super().__init__(min=min, max=max, shape=shape, dtype=dtype, **kwargs)


	def _sample(self, shape):
		return torch.randint(self.min, self.max, shape, generator=self.rng)



class Naturals(Range):
	def __init__(self, max=None, *, min=0, **kwargs):
		super().__init__(min=min, max=max, **kwargs)



class Continuous(Dim):
	def _sample(self, shape):
		return self.unstandardize(self._sample_prior(torch.Size([*shape, *self.shape])))


	def _sample_prior(self, shape):
		return torch.randn(*shape, generator=self.rng)


	def distance(self, x, y, standardize=None):
		N = len(x) if len(x.shape) > 1 else 1
		return self.difference(x, y, standardize=standardize).view(N, -1).norm(p=2, dim=1)



class HalfBound(Continuous):
	def __init__(self, bound=0, side='lower', bound_type='exp', epsilon=1e-10,
	             min=None, max=None, **kwargs):
		
		if min is None and max is None:
			assert bound is not None, 'No bound provided'
			assert side in {'lower', 'upper'}, 'Bound side not specified'
			
			if side == 'lower':
				min = bound
			else:
				max = bound
		assert max is None or min is None, f'Too many bounds specified: min={min}, max={max}'
		assert bound_type in {'exp', 'soft', 'chi', 'abs'}
		if bound_type in {'chi', 'abs'}:
			print(f'WARNING: half-bound dim using transformation {bound_type} cannot be standardized')
		super().__init__(max=max, min=min, **kwargs)
		
		self._bound_type = bound_type
		self._epsilon = epsilon


	def __str__(self):
		terms = ('('+ ', '.join(map(str,self.shape))+'), ') if len(self) > 1 else ''
		lim = f'min={self.min.mean().item():.3g}' \
			if self.min is not None else f'max={self.max.mean().item():.3g}'
		return f'{self.__class__.__name__}({terms}{lim})'


	def standardize(self, vals):
		if self.min is not None:
			vals = vals - self.min.unsqueeze(0)
		elif self.max is not None:
			vals = vals.sub(self.max.unsqueeze(0)).mul(-1)
		
		vals = vals.clamp(min=self._epsilon)
		if self._bound_type == 'soft':
			vals = vals.exp().sub(1).log() # TODO: test this
		elif self._bound_type == 'chi':
			vals = vals.sqrt()
		elif self._bound_type == 'exp':
			vals = vals.log()
		
		return vals


	def unstandardize(self, vals):
		if self._bound_type == 'soft':
			vals = F.softplus(vals)
		elif self._bound_type == 'chi':
			vals = vals.pow(2)
		elif self._bound_type == 'exp':
			vals = vals.exp()
		else:
			vals = vals.abs()
		
		if self.min is not None:
			vals = vals + self.min.unsqueeze(0)
		elif self.max is not None:
			vals = -vals + self.max.unsqueeze(0)
		
		return vals



class Bound(Continuous):
	def __init__(self, min=0., max=1., epsilon=1e-10, **kwargs):
		min, max = torch.as_tensor(min).float(), torch.as_tensor(max).float()
		super().__init__(min=min, max=max, **kwargs)
		assert self.min is not None, f'No lower bound provided'
		assert self.max is not None, f'No upper bound provided'
		
		self._epsilon = epsilon


	def __str__(self):
		terms = ('('+ ', '.join(map(str,self.shape))+'), ') if len(self) > 1 else ''
		return f'{self.__class__.__name__}({terms}min={self.min.mean().item():.3g}, max={self.max.mean().item():.3g})'


	def _sample_prior(self, shape):
		return torch.rand(*shape, generator=self.rng)


	def standardize(self, vals):
		return vals.to(self.min).sub(self.min.unsqueeze(0)).div(self.range.unsqueeze(0)) \
			.clamp(min=self._epsilon, max=1 - self._epsilon).to(vals)


	def unstandardize(self, vals):
		return vals.to(self.range).clamp(min=self._epsilon, max=1 - self._epsilon) \
			.mul(self.range.unsqueeze(0)).add(self.min.unsqueeze(0)).to(vals)


	# def difference(self, x, y, standardize=None):
	# 	return super().difference(x, y, standardize=standardize)# * (self.range.unsqueeze(0) ** float(not standardize))



class Logspace(Bound):
	def standardize(self, vals):
		return super().standardize(vals.log())
	
	
	def unstandardize(self, vals):
		return super().unstandardize(vals).exp()



class Unbound(Continuous):
	def __init__(self, shape=None, min=None, max=None, **kwargs):
		if isinstance(shape, int):
			shape = shape,
		super().__init__(shape=shape, min=None, max=None, **kwargs)


	def standardize(self, vals):
		return vals


	def unstandardize(self, vals):
		return vals



class Periodic(Bound):
	def __init__(self, period=1., min=0., max=None, **kwargs):
		assert min is not None and (period is not None or max is not None), 'Not enough bounds provided'
		if max is None:
			max = min + period
		super().__init__(min=min, max=max, **kwargs)


	def __str__(self):
		return f'{self.__class__.__name__}({self.period.mean().item():.3g})'


	@property
	def period(self):
		return self.range


	@property
	def expanded_shape(self):
		return torch.Size((*self.shape, 2))


	def expand(self, vals):
		thetas = self.standardize(vals).mul(2*np.pi)
		return torch.stack([thetas.cos(), thetas.sin()], -1)


	def compress(self, vals):
		vals = vals.view(-1, *self.expanded_shape)
		return self.unstandardize(torch.atan2(vals[..., 1], vals[..., 0]).div(2 * np.pi).remainder(1))


	def difference(self, x, y, standardize=None):
		if standardize is None or standardize: # by default, standardize
			x, y = self.standardize(x), self.standardize(y)
		return angle_diff(x, y, period=1.) * (self.period ** float(standardize is None))


	def measure(self, x, y, standardize=None):
		if standardize is None or standardize: # by default, standardize
			x, y = self.standardize(x), self.standardize(y)
		return super().measure(x, y)


	def distance(self, x, y, standardize=None):
		return super().distance(x, y, standardize=standardize is None or standardize)


	def transform(self, vals, spec):
		if isinstance(spec, Categorical):
			spec.n += 1
		out = super().transform(vals, spec)
		if isinstance(spec, Categorical):
			spec.n -= 1
		return out



class MultiDim(Dim):
	def __init__(self, channels=None, dim=1, shape=None, **kwargs):
		assert channels is not None or shape is not None, 'need a dimensionality of the topology'
		super().__init__(shape=shape, **kwargs)
		if channels is None:
			channels = (0, *self.shape)[dim]
		assert channels > 0, 'topology must have more than 0 channels'
		self.dim = dim
		self.channels = channels


	def _fingerprint_data(self):
		return {
			'dim': str(self.dim),
			'channels': str(self.channels),
			**super()._fingerprint_data()
		}
	

	def __len__(self):
		return self.channels


	def __str__(self):
		return f'{self.__class__.__name__}({len(self)})'



class Surface(MultiDim):
	def geodesic(self, x, y, standardize=None):
		raise NotImplementedError


	def distance(self, x, y, standardize=None):
		return self.geodesic(x, y, standardize=standardize)



class LebesgueSurface(Surface, Lp):
	def standardize(self, vals):
		return F.normalize(vals, p=self.p, dim=self.dim)


	def __len__(self):
		return self.channels - 1


	def unstandardize(self, vals):
		return self.standardize(vals)


	def magnitude(self, x):
		return super().magnitude(x, dim=self.dim)



class Simplex(LebesgueSurface, Bound, L1):
	def geodesic(self, x, y, standardize=None):
		return super(Surface, self).distance(x, y, standardize=standardize)


	def _sample(self, shape):  # from Donald B. Rubin, The Bayesian bootstrap Ann. Statist. 9, 1981, 130-134.
		# discussed in https://cs.stackexchange.com/questions/3227/uniform-sampling-from-a-simplex

		extra_shape = [*shape, *self.shape]
		raw = super()._sample(shape).view(*extra_shape)

		dim = self.dim
		extra_shape[dim] = 1
		
		edges = torch.cat([torch.zeros(*extra_shape),
		                   raw.narrow(dim, 0, self.channels - 1).sort(dim)[0],
		                   torch.ones(*extra_shape)], dim)
		
		samples = edges.narrow(dim, 1, self.channels) - edges.narrow(dim, 0, self.channels)
		return samples



class Sphere(LebesgueSurface, Unbound, L2):
	def geodesic_cos(self, x, y, standardize=None): # returns cos(theta)
		if standardize:
			x, y = self.standardize(x), self.standardize(y)
		return x.mul(y).sum(self.dim)


	def geodesic(self, x, y, standardize=None): # theta: [0, pi]
		return self.geodesic_cos(x, y, standardize=standardize).acos()#.div(np.pi/2)


	def measure_euclidean(self, x, y, standardize=None):
		return super(Surface, self).distance(x, y, standardize=standardize)



class Spatial(MultiDim): # order of some dimensions is not permutable (channel dim are permutable)
	def __init__(self, channels, *size, shape=None, channel_first=True, **kwargs):
		if isinstance(size, int):
			size = (size,)
		if not len(size):
			size = shape

		shape = (channels,) if size is None else (
			(channels, *size) if channel_first else (*size, channels))

		self._channel_first = channel_first
		super().__init__(channels=channels, shape=shape, **kwargs)


	@property
	def channel_first(self):
		return self._channel_first
	@channel_first.setter
	def channel_first(self, val):
		if val != self.channel_first:
			channels, size = self.channels, self.size
			self.shape = (channels, *size) if val else (*size, channels)
			self._channel_first = val


	@property
	def channels(self):
		return self.shape[0] if self.channel_first else self.shape[-1]
	@channels.setter
	def channels(self, val):
		self.shape = (val, *self.size) if self.channel_first else (*self.size, val)


	@property
	def C(self):
		return self.channels
	@C.setter
	def C(self, val):
		self.channels = val


	@property
	def size(self):
		return self.shape[1:] if self.channel_first else self.shape[:-1]
	@size.setter
	def size(self, val):
		if isinstance(val, int):
			val = (val,)
		self.shape = (self.channels, *val) if self.channel_first else (*val, self.channels)


	def __str__(self):
		return f'{self.__class__.__name__}(C={self.channels}, size={self.size})'



class Sequence(Spatial):
	def __init__(self, channels=1, length=None, **kwargs):
		super().__init__(channels=channels, shape=(length,), **kwargs)
		self._length = length

	@property
	def length(self):
		return self._length
	@length.setter
	def length(self, val):
		self._length = val
		self.size = (val,)


	@property
	def L(self):
		return self.length
	@L.setter
	def L(self, val):
		self.length = val


	def __str__(self):
		return f'{self.__class__.__name__}(C={self.channels}, L={self.length})'



class Image(Spatial):
	def __init__(self, channels=1, height=None, width=None, **kwargs):
		super().__init__(channels=channels, shape=(height, width), **kwargs)
		self._height = height
		self._width = width

	@property
	def height(self):
		return self._height
	@height.setter
	def height(self, val):
		self._height = val
		self.size = (val, self.width)


	@property
	def width(self):
		return self._width
	@width.setter
	def width(self, val):
		self._width = val
		self.size = (self.height, val)

	@property
	def H(self):
		return self.height
	@H.setter
	def H(self, val):
		self.height = val


	@property
	def W(self):
		return self.width
	@W.setter
	def W(self, val):
		self.width = val


	def __str__(self):
		return f'{self.__class__.__name__}(C={self.channels}, H={self.height}, W={self.width})'



class Pixels(Image, Bound):
	def __init__(self, channels=1, height=None, width=None, as_bytes=True, min=None, max=None, dtype=None, **kwargs):
		min, max = (0, 255) if as_bytes else (0., 1.)
		if dtype is None:
			dtype = torch.uint8 if as_bytes else torch.float
		super().__init__(channels=channels, height=height, width=width, min=min, max=max, dtype=dtype, **kwargs)
		self.as_bytes = as_bytes



class Volume(Spatial):
	def __init__(self, channels=1, height=None, width=None, depth=None, **kwargs):
		super().__init__(channels=channels, shape=(height, width, depth), **kwargs)
		self._height = height
		self._width = width
		self._depth = depth


	@property
	def height(self):
		return self._height
	@height.setter
	def height(self, val):
		self._height = val
		self.size = (val, self.width, self.depth)


	@property
	def width(self):
		return self._width
	@width.setter
	def width(self, val):
		self._width = val
		self.size = (self.height, val, self.depth)


	@property
	def depth(self):
		return self._depth
	@depth.setter
	def depth(self, val):
		self._depth = val
		self.size = (self.height, self.width, val)


	@property
	def H(self):
		return self.height
	@H.setter
	def H(self, val):
		self.height = val

	@property
	def W(self):
		return self.width
	@W.setter
	def W(self, val):
		self.width = val

	@property
	def D(self):
		return self.depth
	@D.setter
	def D(self, val):
		self.depth = val

	def __str__(self):
		return f'{self.__class__.__name__}(C={self.channels}, H={self.height}, W={self.width}, D={self.depth})'



class Categorical(Dim):
	def __init__(self, n, **kwargs):
		if isinstance(n, _Sequence):
			n = list(n)
			n, values = len(n), n
		else:
			assert isinstance(n, int), f'bad: {n}'
			n, values = n, list(range(n))
		super().__init__(min=0, max=n - 1, **kwargs)
		# super().__init__(min=torch.as_tensor(0), max=torch.as_tensor(n - 1), **kwargs)
		self._min = self._min#.long()
		self._max = self._max#.long()
		self.n = n
		self.values = values


	@property
	def width(self): # for (default) neural representation
		return self.n


	def validate(self, value):
		if value in self.values:
			return
		return super().validate(value)


	def __str__(self):
		return f'{self.__class__.__name__}({self.n})'


	def _fingerprint_data(self):
		return {'n': str(self.n), **super()._fingerprint_data()}
	

	def standardize(self, vals):
		return vals / (self.n - 1)


	def unstandardize(self, vals):
		return (vals * self.n).long().clamp(max=self.n - 1)


	@property
	def expanded_shape(self):
		return torch.Size((*self.shape, self.n))


	def _sample(self, shape):
		samples = torch.randint(self.n, size=(*shape, *self.shape), generator=self.rng)
		return samples


	def expand(self, vals):
		return F.one_hot(vals.long(), self.n)


	def compress(self, vals):
		return vals.max(-1)[1]


	def difference(self, x, y, standardize=None):
		return (x - y).bool().float()


	def distance(self, x, y, standardize=None):
		N = len(x) if len(x.shape) > 1 else 1
		return self.difference(x, y).view(N, -1).sum(-1).bool().float()



class Selection(Categorical):
	def __init__(self, values=None, **kwargs):
		if values is None:
			values = []
		super().__init__(values, **kwargs)


	def __str__(self):
		return f'{self.__class__.__name__}({", ".join(map(str,self.values))})'

	def __contains__(self, item):
		return item in self.values

	@property
	def n(self):
		return len(self.values)
	@n.setter
	def n(self, val):
		pass

	def replace(self, values):
		self.values = values

	def append(self, value):
		self.values.append(value)

	def extend(self, values):
		self.values.extend(values)

	def remove(self, value):
		self.values.remove(value)

	def clear(self):
		self.values.clear()



class ImplicitSelection(Selection):
	def __init__(self, value_fn=None, **kwargs):
		super().__init__(None, **kwargs)
		self.value_fn = value_fn

	@property
	def values(self):
		if self.value_fn is None:
			return []
		return self.value_fn()
	@values.setter
	def values(self, val):
		pass




class Binary(Categorical):
	def __init__(self, n=None, **kwargs):
		super().__init__(n=2, **kwargs)


	def __str__(self):
		return f'Binary()'



class Joint(Dim):
	def __init__(self, *dims, names=None, shape=None, max=None, min=None, scales=None,
	             # seed=None, gen=None,
	             **kwargs):
		singles = []
		dim_names = []
		for i, d in enumerate(dims):
			if isinstance(d, Joint):
				singles.extend(d.dims)
				if d.names is not None:
					dim_names.extend(d.names)
				elif names is not None:
					dim_names.extend([f'{names[i]}-{di}' for di, dm in enumerate(d.dims)])
			else:
				singles.append(d)
				if names is not None:
					dim_names.append(names[i])
		dims = singles
		shape = (sum(len(dim) for dim in dims),)
		expanded_shape = torch.Size((sum(dim.expanded_shape.numel() for dim in dims),))
		
		super().__init__(shape=shape, min=None, max=None,
		                 # seed=seed, gen=gen,
		                 **kwargs)
		# if gen is not None:
		# 	for dim in dims:
		# 		dim.gen = gen
		# elif seed is not None:
		# 	for dim in dims:
		# 		seed = self.gen_deterministic_seed(seed)
		# 		dim.seed = seed

		self.names = dim_names
		self.dims = dims
		self._expanded_shape = expanded_shape
		self.scales = scales
		self._dim_indices = np.cumsum([0] + [len(dim) for dim in self.dims]).tolist()
		self._dim_expanded_indices = np.cumsum([0] + [dim.expanded_shape.numel() for dim in self.dims]).tolist()
		self._is_dense = any(1 for dim in dims if isinstance(dim, Continuous))


	def __str__(self):
		contents = ', '.join(str(x) for x in self.dims)
		return f'{self.__class__.__name__}({contents})'


	def __iter__(self):
		return iter(self.dims)


	def _fingerprint_data(self):
		return {'dims': self.dims, **super()._fingerprint_data()}
	

	@property
	def expanded_shape(self):
		return self._expanded_shape
	
	
	def select(self, data, idx, dim=1, use_expanded=False):
		inds = self._dim_expanded_indices if use_expanded else self._dim_indices
		start, stop = inds[idx], inds[idx+1]
		sel = [None]*dim + [slice(start, stop)]
		return data[tuple(sel)]
	
	
	def _dispatch(self, method, *vals, use_expanded=False, **base_kwargs):
		outs = []
		idx = 0
		B = None
		for dim in self.dims:
			D = dim.expanded_len() if use_expanded else len(dim)
			args = tuple(v.narrow(-1, idx, D) for v in vals)
			kwargs = base_kwargs.copy()
			
			out = getattr(dim, method)(*args, **kwargs)
			if B is None:
				B = out.size(0) #if len(out.size()) > 1 else 1
			out = out.view(B, -1)
			if self._is_dense:
				out = out.float()
			outs.append(out)
			
			idx += D
		
		return torch.cat(outs, -1)


	def standardize(self, vals):
		return self._dispatch('standardize', vals)


	def unstandardize(self, vals):
		return self._dispatch('unstandardize', vals)


	def expand(self, vals):
		return self._dispatch('expand', vals)


	def compress(self, vals):
		return self._dispatch('compress', vals, use_expanded=True)


	def _sample(self, shape):
		return self._dispatch('_sample', shape=shape)


	def distance(self, x, y, standardize=None, scale=None):
		if standardize is None and self.scales is not None: # specifying scales override intrinsic scales
			standardize = True
		mags = self.measure(x, y, standardize=standardize)
		if scale is None:
			scale = self.scales or torch.ones(mags.shape[-1]).to(mags)
		return (mags @ scale.view(-1, 1)).squeeze(-1)

		diffs = self.difference(x, y, standardize=standardize)
		if self.scales is None:
			return diffs.sum(-1)


	def measure(self, x, y, standardize=None):
		return self._dispatch('measure', x, y, standardize=standardize)


	def difference(self, x, y, standardize=None):
		return self._dispatch('difference', x, y, standardize=standardize)


	def __getitem__(self, item):
		return self.dims[item]








from .contexts import *
from .kits import *


class AbstractMogul(AbstractGig):
	def __iter__(self):
		return self

	def __next__(self):
		raise NotImplementedError

	@property
	def current(self):
		raise NotImplementedError


class AbstractCrawler(AbstractMogul):
	def select(self, decision: 'AbstractDecision', gizmo: str) -> Any:
		raise NotImplementedError


class SimpleFrame(Cached, Gig, MutableKit, AbstractCrawler):
	def __init__(self, owner, base=None):
		if base is None:
			base = {}
		super().__init__()
		self._owner = owner
		# self.update(base)
		self._frame = base # choices that were made

	@property
	def current(self):
		return self._owner.current

	def __next__(self):
		return next(self._owner)

	def __iter__(self):
		return self._owner

	def gizmos(self) -> Iterator[str]:
		yield from filter_duplicates(super().gizmos(), self._owner.gizmos())

	def select(self, decision: 'AbstractDecision', gizmo: str) -> Any:
		if gizmo in self._frame:
			return self._frame[gizmo]
		value = self._owner.select(decision, gizmo)
		self._frame[gizmo] = value
		return value

	def _grab(self, gizmo: str) -> Any:
		return self._owner.grab_from(self, gizmo)



class AbstractDecision(AbstractGadget):
	def choices(self, gizmo: str) -> Iterator[Any]:
		raise NotImplementedError



class SimpleCrawler(AbstractCrawler): # TODO: include all options if there are multiple vendors!
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._current = None
		self._crawl_stack = OrderedDict()
		self._current_base = {}

	# _StackEntry = namedtuple('_StackEntry', 'gizmo decision remaining')
	_SubCrawler = SimpleFrame
	_SubDecision = AbstractDecision

	def select(self, decision: 'AbstractDecision', gizmo: str) -> Any:
		assert isinstance(decision, self._SubDecision), f'Expected {self._SubDecision}, got {decision}'
		if gizmo in self._current_base: # (past) frames lazily get missing base values from current -> can be tricky
			return self._current_base[gizmo]
		assert gizmo not in self._crawl_stack, f'Gizmo {gizmo} already selected'
		options = decision.choices(gizmo)
		self._crawl_stack[gizmo] = options
		return self._crawl_step()

	def _crawl_step(self):
		assert len(self._crawl_stack) > 0, 'Nothing to crawl'

		gizmo, remaining = self._crawl_stack.popitem()
		if gizmo in self._current_base: # removed previous value from cache
			del self._current_base[gizmo]

		value = next(remaining)
		self._current_base[gizmo] = value
		self._crawl_stack[gizmo] = remaining
		return value

	def __next__(self):
		while len(self._crawl_stack) > 0:
			try:
				self._crawl_step()
			except StopIteration:
				pass
			else:
				self._current = self._create_frame()
				return self._current
		raise StopIteration

	def _create_frame(self) -> Iterator[Any]:
		return self._SubCrawler(self, base=self._current_base.copy())

	@property
	def current(self):
		if self._current is None:
			self._current = self._create_frame()
		return self._current

	def grab_from(self, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		if not isinstance(ctx, self._SubCrawler):# and self._current is not None:
			return self.current.grab_from(self.current, gizmo) # traceless delegation to current
		# current frame has not yet loaded this gizmo - it must be grabbed for real
		return super().grab_from(ctx, gizmo)








import numpy as _np
# import torch as _torch


def deviced_fn(fn, device='cuda', out_device='cpu'):
	def _process_arg(t, d=device):
		if isinstance(t, _np.ndarray):
			t = _torch.as_tensor(t)
		if d is not None and isinstance(t, _torch.Tensor):
			t = t.to(d)
		return t
	def _wrapped(*args, **kwargs):
		args = list(map(_process_arg, args))
		kwargs = {key:_process_arg(val) for key, val in kwargs.items()}
		out = fn(*args, **kwargs)
		return _process_arg(out, d=out_device)
	return _wrapped



def combine_dims(tensor, start=1, end=None):
	if end is None:
		end = len(tensor.shape)
	combined_shape = [*tensor.shape[:start], -1, *tensor.shape[end:]]
	return tensor.view(*combined_shape)



def split_dim(tensor, *splits, dim=0):
	split_shape = [*tensor.shape[:dim], *splits, *tensor.shape[dim+1:]]
	return tensor.view(*split_shape)



def swap_dim(tensor, d1=0, d2=1):
	dims = list(range(len(tensor.shape)))
	dims[d1], dims[d2] = dims[d2], dims[d1]
	return tensor.permute(*dims)









from typing import Optional


class InvalidContextError(TypeError):
	pass



class ToolFailedError(Exception):
	def __init__(self, gizmo: str, message: Optional[str] = None):
		if message is None:
			message = f'{gizmo!r}'
		super().__init__(message)
		self.gizmo = gizmo



class MissingGizmoError(ToolFailedError, KeyError):
	def __init__(self, gizmo: str, message: Optional[str] = None):
		if message is None:
			message = gizmo
		super().__init__(gizmo, message)



class MachineError(MissingGizmoError):
	def __init__(self, gizmo: str, label: str, owner, *, message: Optional[str] = None):
		if message is None:
			message = f'Machine {label} (in {owner}) is missing gizmo {gizmo!r}'
		super().__init__(gizmo, message)






import os
import json
# import h5py as hf
# import torch
from pathlib import Path
from omnibelt import unspecified_argument, agnostic, agnosticproperty, md5

from ..structure import spaces
from ..parameters import hparam, submodule
from ..persistent import Rooted
from ..features import Named, Seedable

from .abstract import AbstractDataRouter
from .routers import Observation, Supervised, Labeled, Synthetic, DataCollection
from .top import Dataset, Datastream, Buffer



class Sampledstream(Dataset, Seedable): # Datastream -> Dataset
	stream = submodule(builder='stream')
	n_samples = hparam(required=True, space=spaces.Naturals(), inherit=True)
	sampler_seed = hparam(16283393149723337453, hidden=True)


	def __init__(self, n_samples: int, **kwargs):
		super().__init__(**kwargs)
		self.n_samples = n_samples
		for gizmo in self.stream.gizmos():
			self.register_buffer(gizmo, space=self.stream.space_of(gizmo))


	def _expected_size(self):
		return self.n_samples


	def _title(self):
		return f'{self.stream.title}[{self.n_samples}]'


	def _sample_stream_content(self, n_samples=None):
		# replacing stream with fixed samples
		if n_samples is None:
			n_samples = self.n_samples
		batch = self.stream.batch(n_samples, seed=self.sampler_seed) # optionally sample in smaller batches
		# batch.reset_rng(self.sampler_seed)
		for gizmo in batch.gizmos():
			buffer = self.get_buffer(gizmo)
			data = batch[gizmo]
			buffer.data = data


	def _prepare(self, **kwargs):
		out = super()._prepare( **kwargs)
		self._sample_stream_content()
		return out


	def space_of(self, gizmo: str):
		try:
			return super().space_of(gizmo)
		except self._ToolFailedError:
			return self.stream.space_of(gizmo)



class SimpleSampledStream(Sampledstream):
	_source_cls = None
	_source_hyperparameters = None
	def __init_subclass__(cls, source_cls=None, **kwargs):
		super().__init_subclass__(**kwargs)
		if source_cls is not None:
			cls._source_cls = source_cls
			hparams = []
			for name, value in source_cls.named_hyperparameters():
				setattr(cls, name, value.copy())
				hparams.append(name)
			cls._source_hyperparameters = tuple(hparams)


	@submodule(hidden=True)
	def stream(self):
		return self._source_cls(**{name: getattr(self, name) for name in self._source_hyperparameters})


	@agnostic
	def signatures(self, owner = None):
		if isinstance(self, type):
			yield from self._source_cls.signatures(owner)
		else:
			yield from super().signatures(owner)



class ObservationDataset(Observation, Dataset):
	pass
	# class Batch(Observation, Dataset.Batch):
	# 	pass
	# class View(Observation, Dataset.View):
	# 	pass



class SupervisedDataset(Supervised, ObservationDataset):
	pass
	# class Batch(Supervised, ObservationDataset.Batch):
	# 	pass
	# class View(Supervised, ObservationDataset.View):
	# 	pass



class LabeledDataset(Labeled, SupervisedDataset):
	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.register_material_alias('target', 'label')


	# class Batch(Labeled, SupervisedDataset.Batch):
	# 	pass
	# class View(Labeled, SupervisedDataset.View):
	# 	pass


	# def generate_observation_from_label(self, label, gen=None):
	# 	raise NotImplementedError



# Labeled means there exists a deterministic mapping from labels to observations
# (not including possible subsequent additive noise)



class SyntheticDataset(Synthetic, LabeledDataset):
	_use_mechanisms = True

	def __init__(self, use_mechanisms=None, **kwargs):
		super().__init__(**kwargs)
		if use_mechanisms is not None:
			self._use_mechanisms = use_mechanisms
		self.register_material_alias('mechanism', 'label')
		self.register_material_alias('target', 'mechanism' if self._use_mechanisms else 'label')


	# class Batch(Synthetic, LabeledDataset.Batch):
	# 	@property
	# 	def _distinct_mechanisms(self):
	# 		return self.source._distince_mechanisms
	# class View(Synthetic, LabeledDataset.View):
	# 	@property
	# 	def _distinct_mechanisms(self):
	# 		return self.source._distince_mechanisms


	# def generate_mechanism(self, *shape):
	# 	return self.mechanism_space.sample(*shape, gen=self.gen)
	#
	#
	# def generate_observation_from_label(self, label, gen=None):
	# 	return self.generate_observation_from_mechanism(self.transform_to_mechanisms(label), gen=gen)
	#
	#
	# def generate_observation_from_mechanism(self, mechanism, gen=None):
	# 	raise NotImplementedError
# Synthetic means the mapping is known (and available, usually only for evaluation)
# TODO: separate labels and mechanisms



class RootedRouter(DataCollection, Rooted):
	_dirname = None

	@agnostic
	def _infer_root(self, root=None):
		if root is not None:
			return Path(root)
		return super()._infer_root(root=root) / 'datasets'


	@agnosticproperty
	def root(self):
		dname = self._dirname
		assert dname is not None, 'missing dataset directory name'
		return self._infer_root() / dname


	@agnostic
	def get_aux_path(self):
		root = self.root / 'aux'
		root.mkdir(parents=True, exist_ok=True)
		return root


	# def _find_path(self, dataset_name='', file_name=None, root=None):
	# 	if root is None:
	# 		root = self.root
	# 	*other, dataset_name = dataset_name.split('.')
	# 	if file_name is None:
	# 		file_name = '.'.join(other) if len(other) else self.name
	# 	path = root / f'{file_name}.h5'
	# 	return path, dataset_name
	#
	#
	# _default_hdf_buffer_type = HDFBuffer
	# def register_hdf_buffer(self, name, dataset_name, file_name=None, root=None,
	#                         buffer_type=None, path=None, **kwargs):
	# 	if buffer_type is None:
	# 		buffer_type = self._default_hdf_buffer_type
	# 	if path is None:
	# 		path, dataset_name = self._find_path(dataset_name, file_name=file_name, root=root)
	# 	return self.register_buffer(name, buffer_type=buffer_type, dataset_name=dataset_name, path=path, **kwargs)
	#
	#
	# @staticmethod
	# def create_hdf_dataset(path, dataset_name, data=None, meta=None, dtype=None, shape=None):
	# 	# if file_name is unspecified_argument:
	# 	# 	file_name = 'aux'
	# 	# if path is None:
	# 	# 	path, dataset_name = self._find_path(dataset_name, file_name=file_name, root=root)
	#
	# 	if isinstance(data, torch.Tensor):
	# 		data = data.detach().cpu().numpy()
	# 	with hf.File(path, 'a') as f:
	# 		if data is not None or (dtype is not None and shape is not None):
	# 			f.create_dataset(dataset_name, data=data, dtype=dtype, shape=shape)
	# 		if meta is not None:
	# 			f.attrs[dataset_name] = json.dumps(meta, sort_keys=True)
	# 	return path, dataset_name



class DownloadableRouter(RootedRouter):
	def __init__(self, download=False, **kwargs):
		super().__init__(**kwargs)
		self._auto_download = download

	def _prepare(self, source=None, **kwargs):
		if not self.is_downloaded():
			if self._auto_download:
				self.download()
			else:
				raise self.DatasetNotDownloaded(self.title)
		super()._prepare(source=source, **kwargs)

	@agnostic
	def is_downloaded(self):
		raise NotImplementedError

	@classmethod
	def download(cls, **kwargs):
		raise NotImplementedError


	class DatasetNotDownloaded(FileNotFoundError):
		def __init__(self, name):
			super().__init__(f'use download=True to enable automatic download for {name}.')



class ImageBuffer(Buffer):
	def process_image(self, image):
		if not self.space.as_bytes:
			return image.float().div(255)
		return image


	def _get_from(self, source, key=None):
		return self.process_image(super()._get_from(source, key=key))









from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Generator, Type, Iterable, \
	Iterator, NamedTuple, ContextManager
import types
from omnibelt import split_dict, unspecified_argument, agnosticmethod, OrderedSet, \
	extract_function_signature, method_wrapper, agnostic, agnosticproperty, \
	defaultproperty, autoproperty, referenceproperty, smartproperty, cachedproperty, TrackSmart, Tracer
from omnibelt.nodes import AddressNode, SparseNode, AutoAddressNode


class Specification:
	def __iter__(self):
		return iter(self.emit())

	def include(self, *srcs: Union['Specification', 'Specced', Iterable[Tuple[str, 'Hyperparameter']]]):
		raise NotImplementedError

	def emit(self, recursive=True, flat=True, **kwargs) -> Iterator[Tuple[str, 'Hyperparameter']]:
		raise NotImplementedError



class LazySpecification(Specification):
	def __init__(self, *srcs, **kwargs):
		super().__init__(**kwargs)
		self.srcs = []
		self.include(*srcs)

	def include(self, *srcs: Union['Specced.Specification', 'Specced', Iterator[Tuple[str, 'Hyperparameter']]]):
		for src in srcs:
			if isinstance(src, types.GeneratorType):
				src = list(src)
			self.srcs.append(src)

	def _package(self, key, param, trace, delimiter=None, **kwargs):
		if delimiter is not None:
			history = trace.path
			if len(history):
				terms = [k for k, _ in history] + [key]
				key = delimiter.join(terms)
		return key, param

	def _emit_params(self, params, *, trace, recursive=True, **kwargs):
		for key, param in params:
			item = self._package(key, param, trace=trace, **kwargs)
			if item is not None:
				yield item
				if recursive and isinstance(param, Specced): # recursive
					yield from self._emit_src(param, trace=trace.append((key, param)), **kwargs)

	def _emit_src(self, src, **kwargs):
		if isinstance(src, Specification):
			yield from self._emit_params(src.emit(**kwargs), **kwargs)
		elif isinstance(src, Specced):
			yield from self._emit_params(src.full_spec().emit(**kwargs), **kwargs)
		else:
			yield from self._emit_params(src, **kwargs)

	Tracer = Tracer

	def emit(self, recursive=True, flat=True, trace=None, delimiter=None,
	         **kwargs) -> Iterator[Tuple[str, 'Hyperparameter']]:
		if flat and delimiter is None:
			delimiter = '.'
		if trace is None:
			trace = self.Tracer()
		for src in self.srcs:
			yield from self._emit_src(src, trace=trace, recursive=recursive, delimiter=delimiter, **kwargs)



class SpecNode(Specification, AutoAddressNode, SparseNode):
	def _include_param(self, key, param, trace=None, recursive=True, **kwargs):
		self.set(key, param)
		if recursive and isinstance(param, Specced): # recursive
			self._include_src(param, trace=trace.append((key, param)), **kwargs)

	Tracer = Tracer
	def _include_src(self, src, trace=None, **kwargs):
		if trace is None:
			trace = self.Tracer()
		if isinstance(src, Specification):
			for key, param in src.emit(**kwargs):
				self._include_param(key, param, trace=trace, **kwargs)
		elif isinstance(src, Specced):
			for key, param in src.full_spec().emit(**kwargs):
				self._include_param(key, param, trace=trace, **kwargs)
		else:
			for key, param in src:
				self._include_param(key, param, trace=trace, **kwargs)

	def include(self, *srcs: Union['Specification', 'Specced', Iterable[Tuple[str, 'Hyperparameter']]]):
		for src in srcs:
			if isinstance(src, types.GeneratorType):
				src = list(src)
			self.add(src)

	def emit(self, recursive=True, flat=True, prefix=None, **kwargs) -> Iterator[Tuple[str, 'Hyperparameter']]:
		if prefix is None:
			prefix = self.Tracer()
		for key, child in self.named_children():
			addr = prefix.append(key)
			if child.has_payload:
				name = self._address_delimiter.join(addr.path) if flat else key
				yield name, child.payload
			if recursive and not child.is_leaf:
				yield from child.emit(recursive=recursive, flat=flat, prefix=addr, **kwargs)


class Specced:
	Specification = SpecNode
	# class Specification(SpecNode): pass

	@agnostic
	def full_spec(self, spec: Optional[Specification] = None) -> Specification:
		if spec is None:
			spec = self.Specification()
		return spec




















from .abstract import AbstractModel, AbstractTrainableModel, AbstractTrainer
from .simple import SimpleModel, Fitable, Evaluatable
from .iterative import TrainableModel, Trainer

# from .estimators import ScikitEstimator, ScikitEstimatorWrapper, Classifier, Regressor, Periodized, \
# 	JointEstimator, ParallelEstimator, MultiEstimator, AbstractScikitEstimator



from .errors import *



class ToolBase(AbstractGadget):
	_ToolFailedError = GadgetFailedError
	_MissingGizmoError = MissingGizmoError


	# def _grab_from(self, ctx: AbstractContext, gizmo: str) -> Any:
	# 	raise NotImplementedError


	# def grab_from(self, ctx: Optional[AbstractContext], gizmo: str,
	#               default: Optional[Any] = unspecified_argument) -> Any:
	# 	try:
	# 		return self._grab_from(ctx, gizmo)
	# 	except self._ToolFailedError:
	# 		if default is unspecified_argument:
	# 			raise
	# 		return default


class SingleGizmoTool(ToolBase):
	def __init__(self, gizmo: str, **kwargs):
		super().__init__(**kwargs)
		self._gizmo = gizmo

	def gizmos(self) -> Iterator[str]:
		yield self._gizmo

	def gives(self, gizmo: str) -> bool:
		return gizmo == self._gizmo


class FunctionTool(SingleGizmoTool):
	def __init__(self, gizmo: str, fn: Callable, **kwargs):
		super().__init__(gizmo=gizmo, **kwargs)
		self._fn = fn


	def __repr__(self):
		name = getattr(self._fn, '__qualname__', None)
		if name is None:
			name = getattr(self._fn, '__name__', None)
		return f'{self.__class__.__name__}({name}: {self._gizmo})'


	@property
	def __call__(self):
		return self._fn#.__get__(None, type(None))


	def __get__(self, instance, owner):
		return self._fn.__get__(instance, owner)
		if instance is None:
			return self
		return self._fn.__get__(instance, owner)


	@staticmethod
	def _extract_gizmo_args(fn: Callable, ctx: AbstractGig,
							args: Optional[Tuple] = None, kwargs: Optional[Dict[str, Any]] = None) \
			-> Tuple[Tuple, Dict[str, Any]]:
		return extract_function_signature(fn, default_fn=lambda gizmo, default: ctx.grab_from(ctx, gizmo))


	def grab_from(self, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		if gizmo != self._gizmo:
			raise self._MissingGizmoError(gizmo)
		return self._fn(ctx)



class AutoFunctionTool(FunctionTool):
	@staticmethod
	def _extract_gizmo_args(fn: Callable, ctx: AbstractGig,
							args: Optional[Tuple] = None, kwargs: Optional[Dict[str, Any]] = None) \
			-> Tuple[Tuple, Dict[str, Any]]:
		return extract_function_signature(fn, default_fn=lambda gizmo, default: ctx.grab_from(ctx, gizmo))


	def grab_from(self, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		if gizmo != self._gizmo:
			raise self._MissingGizmoError(gizmo)

		args, kwargs = self._extract_gizmo_args(self._fn, ctx)
		return self._fn(*args, **kwargs)



class ToolSkill(AbstractSkill):
	def __init__(self, *, unbound_fn: Callable, base: Optional[AbstractCraft] = None, **kwargs):
		super().__init__(**kwargs)
		self._unbound_fn = unbound_fn
		self._base = base
		

	@property
	def __call__(self):
		return self._unbound_fn
	

	def __get__(self, instance, owner):
		if instance is None:
			return self
		return self._unbound_fn.__get__(instance, owner)




class FunctionSkill(FunctionTool, ToolSkill):
	pass


class AutoFunctionSkill(AutoFunctionTool, ToolSkill):
	pass


class ToolCraft(FunctionTool, NestableCraft):
	def _wrapped_content(self): # wrapped method
		return self._fn


	_ToolSkill = FunctionSkill
	def as_skill(self, owner: AbstractCrafty) -> ToolSkill:
		unbound_fn = self._wrapped_content_leaf()
		fn_name = getattr(unbound_fn, '__name__', None)
		if fn_name is None:
			fn = unbound_fn.__get__(owner, type(owner))
		else:
			fn = getattr(owner, fn_name) # get most recent version of method
		return self._ToolSkill(self._gizmo, fn=fn, unbound_fn=unbound_fn, base=self)



class AutoFunctionToolCraft(AutoFunctionTool, NestableCraft):
	def _wrapped_content(self): # wrapped method
		return self._fn


	_ToolSkill = AutoFunctionSkill
	def as_skill(self, owner: AbstractCrafty):
		unbound_fn = self._wrapped_content_leaf()
		fn_name = getattr(unbound_fn, '__name__', None)
		if fn_name is None:
			fn = unbound_fn.__get__(owner, type(owner))
		else:
			fn = getattr(owner, fn_name) # get most recent version of method
		return self._ToolSkill(self._gizmo, fn=fn, unbound_fn=unbound_fn, base=self)



class ToolDecorator(ToolBase):
	def __init__(self, gizmo: str, **kwargs):
		super().__init__(**kwargs)
		self._gizmo = gizmo


	def gizmos(self) -> Iterator[str]:
		yield self._gizmo


	def gives(self, gizmo: str) -> bool:
		return gizmo == self._gizmo


	def grab_from(self, ctx: Optional['AbstractGig'], gizmo: str) -> Any:
		raise self._ToolFailedError(gizmo)


	_ToolCraft = ToolCraft
	def _actualize_tool(self, fn: Callable, **kwargs):
		return self._ToolCraft(self._gizmo, fn=fn, **kwargs)


	def __call__(self, fn):
		return self._actualize_tool(fn)



class AutoToolDecorator(ToolDecorator):
	_ToolCraft = AutoFunctionToolCraft





# from .industrial import IndustrialBase
# from omnidata.tools.old.materials import material, Material
from .crafts import SignatureCraft
from .top import Guru, Context, Industrial, Spatial, material, space, indicator, machine, Signature




from .simple import Named, Prepared, ProgressBarred, InitWall
from .containers import SourceContainer, ScoreContainer, Container
from .hardware import Device, DeviceContainer
from .random import RNGManager, Seeded, Seedable, force_rng, \
	gen_deterministic_seed, gen_random_seed, create_rng, default_rng



from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable
from collections import OrderedDict
# import torch

from ..features import Seeded, Prepared
from ..tools.abstract import AbstractScope
from ..tools.context import SizedContext, Cached

from .abstract import AbstractView, AbstractRouterView, AbstractBatch, AbstractDataRouter, AbstractIndexedData, \
	AbstractProgression, AbstractBatchable, AbstractSelector, AbstractCountableData, AbstractCountableRouterView



class ViewBase(AbstractView):
	def __init__(self, source: AbstractDataRouter = None, **kwargs):
		super().__init__(source=source, **kwargs)
		self._source = source


	@property
	def source(self):
		return self._source



class RouterViewBase(ViewBase, AbstractRouterView):
	pass



class SizeSelector(SizedContext, RouterViewBase, AbstractSelector, AbstractCountableData):
	def compose(self, other: AbstractSelector) -> AbstractSelector:
		if isinstance(other, SizeSelector):
			self._size = min(self._size, other._size)
		return self



class SingleIndexSelector(SizeSelector):
	def __init__(self, index: int, **kwargs):
		super().__init__(size=1, **kwargs)
		self._index = index


	@property
	def index(self):
		return self._index


	@property
	def size(self):
		return 1



class IndexView(RouterViewBase, AbstractIndexedData):
	def __init__(self, indices, **kwargs):
		super().__init__(indices=indices, **kwargs)
		self._indices = indices


	@property
	def indices(self):
		return self._indices


# print()
# print('\n'.join(map(str, IndexView.mro())))
# print()
# print('\n'.join(map(str, SizeSelector.mro())))
class IndexSelector(IndexView, SizeSelector):
	def compose(self, other: AbstractSelector) -> AbstractSelector:
		# TODO: check case where other is just the indices directly
		if isinstance(other, IndexSelector) and other.indices is not None:
			self._indices = other.indices[self.indices]
		return self


	@property
	def size(self):
		if self._size is None:
			return len(self.indices)
		return self._size

# class IndexView(IndexSelector, AbstractCountableRouterView): # -> Subset
# 	def validate_context(self, selection: 'AbstractSelector'):
# 		return super().validate_context(selection.compose(self))



# class CachedView(AbstractRouterView):
# 	def __init__(self, source=None, cache_table=None, **kwargs):
# 		if cache_table is None:
# 			cache_table = self._CacheTable()
# 		super().__init__(source=source, **kwargs)
# 		self._cache_table = cache_table
#
# 	_CacheTable = OrderedDict
#
# 	def cached(self):
# 		yield from self._cache_table.keys()
#
# 	def clear_cache(self):
# 		self._cache_table.clear()
#
# 	def __str__(self):
# 		cached = set(self.cached())
# 		return f'{self._title()}(' \
# 		       f'{", ".join((key if key in cached else "{" + key + "}") for key in self.available_buffers())})'
#
# 	def _get_from(self, source, key=None):
# 		if key not in self._cache_table:
# 			out = super()._get_from(self, key)
# 			self._cache_table[key] = out
# 		if source is None or source is self:
# 			return self._cache_table[key]
# 		return self._cache_table[key][source.indices]




# class MissingIndicesError(ValueError): pass

# @property
# def indices(self):
# 	if self._indices is None:
# 		raise self.MissingIndicesError
# 	return self._indices
#
# @property
# def size(self):
# 	if self._size is None:
# 		return len(self.indices)
# 	return self._size

# def compose(self, other: 'IndexBatch') -> 'IndexBatch':
# 	return self.indices[other.indices]








from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable

import math
# import torch

from omniplex.features import Prepared, ProgressBarred
from omniplex.tools.abstract import AbstractScope, AbstractTool, AbstractResource
from omniplex.tools.moguls import BatchMogul, IteratorMogul, SelectionMogul, LimitMogul, \
	BatchBudgetStatMogul, EpochStatMogul, EpochBudgetMogul, SimpleMogul

from omniplex.data.abstract import AbstractProgression, AbstractContext
from omniplex.data.errors import BudgetExceeded, EpochEnd, UnknownSize
from omniplex.data.sources import Shufflable



# class AbstractBudgetProgression(AbstractProgression, BatchBudgetStatMogul):
# 	pass


_inf = float('inf')

class BudgetProgression(BatchProgression, ProgressionBase, BatchBudgetStatMogul):
	def __init__(self, *, sample_limit=None, batch_limit=None,
	             strict_limit=True, strict_batch_size=False, **kwargs):
		super().__init__(**kwargs)
		self._sample_limit = sample_limit
		self._batch_limit = batch_limit

		self._strict_limit = strict_limit
		self._strict_batch_size = strict_batch_size

		self._total_samples = None
		self._total_batches = None


	_BudgetExceeded = BudgetExceeded
	def _validate_context(self, ctx: AbstractContext) -> AbstractContext:
		ctx = super()._validate_context(ctx)
		if self.remaining_samples < 0:
			raise self._BudgetExceeded(f'Exceeded sample budget of {self._total_samples} samples')
		if self.remaining_batches < 0:
			raise self._BudgetExceeded(f'Exceeded batch budget of {self._total_batches} batches')
		return ctx


	def _create_context(self, *args, size=None, **kwargs):
		if size is None:
			size = self._batch_size if self._strict_batch_size else min(self._batch_size, self.remaining_samples)
		return super()._create_context(*args, size=size, **kwargs)


	def _prepare(self, *args, **kwargs):
		super()._prepare(*args, **kwargs)
		self._total_samples, self._total_batches = self.compute_budget(
			samples_per_batch=self.batch_size, strict_batch_size=self._strict_batch_size,
			sample_limit=self._sample_limit, batch_limit=self._batch_limit, strict_limit=self._strict_limit,
		)


	@property
	def budget_samples(self) -> Optional[int]:
		return self._total_samples
	@property
	def budget_batches(self) -> Optional[int]:
		return self._total_batches


	@property
	def remaining_samples(self):
		if self._total_samples is None:
			return _inf
		return self._total_samples - self.sample_count
	@property
	def remaining_batches(self):
		if self._total_batches is None:
			return _inf
		return self._total_batches - self.batch_count


	def done(self):
		return self.remaining_samples <= 0 or self.remaining_batches <= 0



class InfiniteProgression(EpochProgression):
	def __init__(self, infinite=False, **kwargs):
		super().__init__(**kwargs)
		self._infinite = infinite
		self._completed_epochs = 0


	def _epoch_end(self):
		if self._infinite:
			self._completed_epochs += 1
			self._indices = None
			return self._next_batch_indices(self.batch_size)
		else:
			raise self._EpochEnd


	@property
	def current_epoch(self) -> int:
		return self.completed_epochs + 1


	@property
	def completed_epochs(self) -> int:
		return self._completed_epochs



class EpochBudgetProgression(InfiniteProgression, ShuffleProgression, BudgetProgression,
                             BatchBudgetStatMogul, EpochBudgetMogul):
	def __init__(self, batch_size, *, epochs=None, **kwargs):
		super().__init__(batch_size=batch_size, **kwargs)
		self._epochs = epochs
		self._full_epochs = None


	def _prepare(self, *args, **kwargs):
		super()._prepare(*args, **kwargs)

		if self._epochs is None \
				and self._sample_limit is None \
				and self._batch_limit is None \
				and not self._infinite:
			self._epochs = 1

		self._total_samples, self._total_batches, self._full_epochs = self.compute_epoch_budget(
			dataset_size=self.epoch_size, samples_per_batch=self.batch_size,
			strict_batch_size=self._strict_batch_size,
			epochs=self._epochs, sample_limit=self._sample_limit,
			batch_limit=self._batch_limit, strict_limit=self._strict_limit
		)


	def _epoch_end(self):
		self._completed_epochs += 1
		self._setup_epoch()


	def _generate_sample_order(self):
		indices = super()._generate_sample_order()
		if self._strict_limit:
			indices = indices[:self.remaining_samples]
		return indices


	@property
	def full_epochs(self) -> Optional[int]:
		return self._full_epochs


	@property
	def remaining_epochs(self):
		if self._full_epochs is None:
			return _inf
		return self._full_epochs - self.current_epoch



# class BarredProgression(ProgressionBase, ProgressBarred):
# 	def __init__(self, batch_size, *, use_pbar=False, pbar_samples=True, **kwargs):
# 		super().__init__(batch_size=batch_size, **kwargs)
# 		self._use_pbar = use_pbar
# 		self._pbar_samples = pbar_samples
#
#
# 	def _create_pbar(self, *, unit=None, **kwargs):
# 		if unit is None:
# 			unit = 'smpl' if self._pbar_samples else 'batch'
# 		return super()._create_pbar(unit=unit, **kwargs)
#
#
# 	def _prepare(self, *args, **kwargs):
# 		super()._prepare(*args, **kwargs)
# 		if self._use_pbar:
# 			self._create_pbar()
#
#
# 	def _create_batch(self):
# 		pbar = self._pbar
# 		try:
# 			batch = super()._create_batch()
# 		except StopIteration:
# 			if pbar is not None:
# 				pbar.close()
# 			raise
# 		else:
# 			if pbar is not None:
# 				pbar.update(self.batch_size if self._pbar_samples else 1)
# 			return batch
#
#
# 	def set_description(self, desc):
# 		if self._pbar is not None:
# 			self._pbar.set_description(desc)
#
#
#
# class TrackedProgression(BarredProgression, AbstractBudgetProgression): # TODO: add a progress bar
# 	def _create_pbar(self, total=None, **kwargs):
# 		if total is None:
# 			total = self.budget_samples if self._pbar_samples else self.budget_batches
# 		return super()._create_pbar(total=total, **kwargs)



class StreamProgression(BudgetProgression, BatchProgression):
	pass



class SetProgression(EpochBudgetProgression, EpochProgression):
	pass













from typing import Optional, Type

from omnibelt import unspecified_argument, filter_duplicates

# from ..parameters import hparam, with_hparams, Parameterized
from ..tools.top import Cached, DynamicContext, SeededContext, ScopedContext

from .abstract import AbstractBatchable, AbstractCountableData, AbstractBatch
from .views import ViewBase, SizeSelector, IndexSelector
from .progression import AbstractProgression, StreamProgression, SetProgression



class Batchable(AbstractBatchable):
	# _Batch: Type[AbstractBatch] = None
	_Progression: Type[AbstractProgression] = None


	def iterate(self, batch_size: Optional[int] = unspecified_argument, **kwargs):
		self.prepare()
		if batch_size is not unspecified_argument:
			kwargs['batch_size'] = batch_size
		return self._Progression(**kwargs).set_source(self)



class BatchableView(Batchable, ViewBase):
	def iterate(self, batch_size: Optional[int] = unspecified_argument, **kwargs):
		self.prepare()
		if batch_size is not unspecified_argument:
			kwargs['batch_size'] = batch_size
		if self._Progression is None:
			return self.source._Progression(self, **kwargs).set_source(self)
		return self._Progression(**kwargs).set_source(self)



## top level



class BatchBase(Cached, SizeSelector, ScopedContext, DynamicContext, BatchableView, AbstractBatch):
	def __init__(self, progress: AbstractProgression = None, **kwargs):
		super().__init__(progress=progress, **kwargs)
		self._progress = progress


	def sources(self):
		yield from filter_duplicates((self.source, ), super().sources())


	def set_progress(self, progress: AbstractProgression):
		self._progress = progress


	@property
	def progress(self):
		return self._progress

	# def __init__(self, progress: AbstractProgression = None, **kwargs):
	# 	super().__init__(progress=progress, **kwargs)
	# 	self._progress = progress
	#
	#
	# @property
	# def source(self):
	# 	if self._source is None:
	# 		return self.progress.source
	# 	return self._source
	#
	#
	# @property
	# def progress(self):
	# 	return self._progress


	def __str__(self):
		gizmos = [(gizmo if self.is_cached(gizmo) else '{' + gizmo + '}') for gizmo in self.gizmos()]
		return f'{self.title}({", ".join(gizmos)})'



class IndexBatch(IndexSelector, BatchBase):
	pass
	# def __init__(self, *args, **kwargs):
	# 	super().__init__(*args, **kwargs)
	# 	if self.source is not None:
	# 		self.include(self.source)





# class Batchable(BatchableBase, Parameterized):
# 	_Progression = StreamProgression
#
# 	batch_size = hparam(inherit=True)
#
# 	sample_limit = hparam(inherit=True)
# 	batch_limit = hparam(inherit=True)
#
# 	strict_limit = hparam(False, inherit=True, hidden=True)
# 	strict_batch_size = hparam(False, inherit=True, hidden=True)
#
# 	use_pbar = hparam(False, inherit=True, hidden=True)
# 	pbar_samples = hparam(True, inherit=True, hidden=True)
#
#
# 	@with_hparams
# 	def iterate(self, batch_size, sample_limit=None, batch_limit=None,
# 	            strict_limit=False, strict_batch_size=False,
# 	            use_pbar=False, pbar_samples=True, **kwargs):
# 		return super().iterate(batch_size=batch_size, sample_limit=sample_limit, batch_limit=batch_limit,
# 		                       strict_limit=strict_limit, strict_batch_size=strict_batch_size,
# 		                       use_pbar=use_pbar, pbar_samples=pbar_samples, **kwargs)
#
#
#
# class Epochable(Batchable, AbstractCountableData):
# 	_Progression = SetProgression
#
# 	epoch_limit = hparam(inherit=True)
# 	shuffle_batches = hparam(False, inherit=True, hidden=True)
# 	infinite_iteration = hparam(False, inherit=True, hidden=True)
#
# 	@with_hparams
# 	def iterate(self, batch_size, epochs=None, epoch_limit=None, shuffle=None, shuffle_batches=False,
# 	            infinite_iteration=False, infinite=None, **kwargs):
# 		if epoch_limit is None:
# 			epoch_limit = epochs
# 		if shuffle is None:
# 			shuffle = shuffle_batches
# 		if infinite is None:
# 			infinite = infinite_iteration
# 		return super().iterate(batch_size=batch_size, epochs=epoch_limit, shuffle=shuffle,
# 		                       infinite=infinite, **kwargs)



























from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Type, Iterable, Iterator
from pathlib import Path
from omnibelt import agnosticmethod
from omnibelt.exporting import ExportManager, Exporter, Exportable, export, load_export
from omnibelt import exporting_common as _exporting_common

import numpy as np
import pandas as pd
# import h5py as hf
# import torch
# from PIL import Image

from .filesystem import Rooted


class NumpyExport(Exporter, extensions='.npy'):
	@staticmethod
	def validate_export_obj(obj: Any) -> bool:
		return isinstance(obj, np.ndarray)
	@staticmethod
	def _load_export(path: Path, src: Type['ExportManager'], **kwargs) -> Any:
		return np.load(path, **kwargs)
	@staticmethod
	def _export_self(payload: Any, path: Path, src: Type['ExportManager'], **kwargs) -> Optional[Path]:
		return np.save(path, payload, **kwargs)



class NpzExport(Exporter, extensions='.npz'):
	@staticmethod
	def _load_export(path: Path, src: Type['ExportManager'], auto_load=False, **kwargs) -> Any:
		obj = np.load(path, **kwargs)
		if auto_load:
			return {key: obj[key] for key in obj.keys()}
		return obj
	@staticmethod
	def _export_payload(payload: Any, path: Path, src: Type['ExportManager']) -> Optional[Path]:
		args, kwargs = ([], payload) if isinstance(payload, dict) else (payload, {})
		return np.savez(path, *args, **kwargs)



class PandasExport(Exporter, extensions='.csv', types=pd.DataFrame):
	@staticmethod
	def _load_export(path: Path, src: Type['ExportManager'], **kwargs) -> Any:
		return pd.read_csv(path, **kwargs)
	@staticmethod
	def _export_payload(payload: Any, path: Path, src: Type['ExportManager'], **kwargs) -> Optional[Path]:
		return payload.to_csv(path, **kwargs)
	
	

class ImageExport(Exporter, extensions='.png'):
	@staticmethod
	def validate_export_obj(obj, **kwargs):
		return isinstance(obj, Image.Image)
	@staticmethod
	def _load_export(path: Path, src: Type['ExportManager'], **kwargs) -> Any:
		return Image.open(path, **kwargs)
	@staticmethod
	def _export_payload(payload: Any, path: Path, src: Type['ExportManager'], **kwargs) -> Optional[Path]:
		return payload.save(path, **kwargs)



class NumpyImageExport(ImageExport, extensions='.png'):
	@staticmethod
	def _export_payload(payload: Any, path: Path, src: Type['ExportManager'], **kwargs) -> Optional[Path]:
		if not isinstance(payload, Image.Image):
			if payload.dtype != np.uint8:
				raise NotImplementedError
			if len(payload.shape) == 3:
				if payload.shape[0] in {1, 3, 4} and payload.shape[2] not in {1, 3, 4}:
					payload = payload.transpose(1, 2, 0)
				if payload.shape[2] == 1:
					payload = payload.reshape(*payload.shape[:2])
			payload = Image.fromarray(payload)
		return super()._export_self(payload, path, src=src, **kwargs)



class JpgExport(ImageExport, extensions='.jpg'):
	pass



class HDFExport(Exporter, extensions=['.h5', '.hf', '.hdf']):
	@staticmethod
	def _load_export(path: Path, src: Type['ExportManager'], *, auto_load=False, mode='r', **kwargs) -> Any:
		f = hf.File(str(path), mode=mode, **kwargs)
		if auto_load:
			obj = {k: f[k][()] for k in f.keys()}
			f.close()
		else:
			obj = f
		return obj
	@staticmethod
	def _export_payload(payload: Any, path: Path, src: Type['ExportManager'], *,
	                    mode='a', meta=None, **kwargs) -> Optional[Path]:
		with hf.File(str(path), mode=mode, **kwargs) as f:
			if meta is not None:
				for k, v in meta.items():
					f.attrs[k] = v
			for k, v in payload.items():
				f.create_dataset(name=k, data=v)
		return path



class DictExport(Exporter, extensions='', types=dict):
	@staticmethod
	def validate_export_obj(obj, **kwargs):
		return isinstance(obj, dict) and all(isinstance(key, str) for key in obj)
	@staticmethod
	def _load_export(path: Path, src: Type['ExportManager']) -> Any:
		return {item.stem: src.load_export(path=item) for item in path.glob('*')}
	@staticmethod
	def _export_payload(payload: Any, path: Path, src: Type['ExportManager']) -> Optional[Path]:
		path.mkdir(exist_ok=True)
		for key, val in payload.items():
			src.export(val, name=key, root=path)
		return path



class PytorchExport(Exporter, extensions=['.pt', '.pth.tar'], head=True):
	@staticmethod
	def validate_export_obj(obj, **kwargs):
		return True
	@staticmethod
	def _load_export(path: Path, src: Type['ExportManager'], **kwargs) -> Any:
		return torch.load(path, **kwargs)
	@staticmethod
	def _export_payload(payload: Any, path: Path, src: Type['ExportManager'], **kwargs) -> Optional[Path]:
		return torch.save(payload, path, **kwargs)






from omnibelt.crafts import HiddenCrafty

from ..tools import Industrial, SignatureCraft, Signature
from .abstract import AbstractArgumentBuilder
from .hyperparameters import InheritableHyperparameter
from .parameterized import ModifiableParameterized, FingerprintedParameterized, InheritHparamsDecorator, \
	InheritableParameterized, HparamWrapper, SpatialParameterized
from .building import ConfigBuilder, BuilderBase, MultiBuilderBase, RegistryBuilderBase, \
	HierarchyBuilderBase, ModifiableProduct, AnalysisBuilder, RegisteredProductBase
from .submodules import SubmoduleBase, SubmachineBase
from .spec import ArchitectBase, Specced, SpecBase, PlannedBase
# from .spec import PreparedParameterized, SpeccedBase, BuilderSpecced, StatusSpec, BuildableSpec



class hparam(InheritableHyperparameter):
	pass



class submodule(SubmoduleBase):
	pass



class submachine(SubmachineBase):
	pass



class inherit_hparams(InheritHparamsDecorator):
	pass



class with_hparam(HparamWrapper):
	pass


# class Hyperparameter(InheritableHyperparameter, ConfigHyperparameter):
# 	pass
#
#
# class Submodule(Hyperparameter, SubmoduleBase):
# 	pass


# class Spec(StatusSpec, BuildableSpec):
# 	# TODO: spec -> config (and config -> spec (?))
# 	pass




class Parameterized(SpatialParameterized, ModifiableParameterized,
                    InheritableParameterized, FingerprintedParameterized):
	pass



class Spec(SpecBase):
	pass
PlannedBase._Spec = Spec



# class Parameterized(SpeccedBase, ModifiableParameterized, FingerprintedParameterized, PreparedParameterized):
class Structured(Specced, Parameterized):
	@classmethod
	def inherit_hparams(cls, *names):
		out = super().inherit_hparams(*names)
		for name in names:
			val = getattr(cls, name, None)
			if isinstance(val, submachine) and len(cls._inherited_tool_relabels):
				setattr(cls, name, val.replace(cls._inherited_tool_relabels))
		return out



class Function(Structured, HiddenCrafty):
	_auto_machine = SignatureCraft
	def __init_subclass__(cls, signature: Signature = None, **kwargs):
		super().__init_subclass__(**kwargs)
		if signature is not None:
			if signature.fn is None:
				signature.fn = getattr(cls, '__call__', None)
			cls._hidden_crafts.append(cls._auto_machine(signature))



class SimpleFunction(Function):
	_Signature = Signature
	def __init_subclass__(cls, inputs=None, output=None, metas=None, signature=None, **kwargs):
		if signature is None:
			signature = cls._Signature(inputs=inputs, output=output, metas=metas)
		super().__init_subclass__(signature=signature, **kwargs)



# not recommended as it can't handle modifiers



# class BasicBuilder(ConfigBuilder, BuilderSpecced, Parameterized): # AutoBuilder
# 	pass



class MatchingBuilder(Structured, AbstractArgumentBuilder):
	'''Automatically fills in common hyperparameters between the builder and the product'''
	fillin_hparams = hparam(True, inherit=True, hidden=True)
	fillin_hidden_hparams = hparam(False, inherit=True, hidden=True)


	def _matching_hparams(self, product):
		known = set(key for key, _ in self.named_hyperparameters())
		for key, _ in product.named_hyperparameters(hidden=self.fillin_hidden_hparams):
			if key in known:
				yield key


	def _build_kwargs(self, product, *args, **kwargs):
		kwargs = super()._build_kwargs(product, *args, **kwargs)
		if self.fillin_hparams and issubclass(product, Parameterized):
			for key in self._matching_hparams(product):
				if key not in kwargs:
					try:
						val = getattr(self, key)
					except AttributeError:
						continue
					else:
						kwargs[key] = val
		return kwargs



class Builder(MatchingBuilder, ArchitectBase, ModifiableProduct, AnalysisBuilder):#(ConfigBuilder, Parameterized):
	#, inheritable_auto_methods=['product_base']):
	pass



# class Buildable(Builder, BuildableBase):
# 	pass



class MultiBuilder(MultiBuilderBase, Builder):#, wrap_existing=True):
	pass



class RegistryBuilder(RegistryBuilderBase, Builder, create_registry=False):
	pass



class HierarchyBuilder(HierarchyBuilderBase, RegistryBuilder, create_registry=False):
	pass



class RegisteredProduct(Structured, RegisteredProductBase):
	pass















from .imports import *


T = TypeVar('T')



class AbstractQuirk:
	def __set_name__(self, owner, name):
		raise NotImplementedError


	def __get__(self, instance, owner):
		raise NotImplementedError


	def __set__(self, instance, value):
		raise NotImplementedError


	def __delete__(self, instance):
		raise NotImplementedError



class DescriptorQuirk(AbstractQuirk):
	_attrname = None

	def __str__(self):
		return f'<quirk:{self._attrname}>'


	def __repr__(self):
		return f'{self.__class__.__qualname__}({self._attrname})'


	def __set_name__(self, owner: Type[T], name: str):
		if self._attrname is None:
			self._attrname = name
		elif name != self._attrname:
			raise TypeError(
				f"Cannot assign the same {self.__class__.__qualname__} to two different names "
				f"({self._attrname!r} and {name!r})."
			)
		return self


	def __get__(self, instance: T, owner: Optional[Type[T]] = None) -> Any:
		return self.resolve(instance, owner=owner)


	def __set__(self, instance: T, value: Any):
		self.reset(instance, value)


	def __delete__(self, instance: T):
		self.rip(instance)












class ExtractionError(ValueError):
	def __init__(self, obj):
		super().__init__(obj)
		self.obj = obj



class UnknownObjectError(TypeError):
	pass



class NoObjectError(AttributeError):
	pass















from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Generator, Type, Iterable, \
	Iterator, NamedTuple, ContextManager
import logging
from functools import cached_property
from omnibelt import unspecified_argument, defaultproperty, autoproperty, referenceproperty

from .abstract import AbstractHyperparameter
from .errors import MissingValueError

prt = logging.Logger('Hyperparameters')
ch = logging.StreamHandler()
ch.setFormatter(logging.Formatter('%(levelname)s: %(msg)s'))
ch.setLevel(0)
prt.addHandler(ch)


class IgnoreUpdateFlag(ValueError):
	pass



class DefaultProperty:
	_unknown = object()

	def __init__(self, default=unspecified_argument, *, fget=None, fset=None, fdel=None, fformat=None, cache=True,
	             attrname=None, **kwargs):
		if default is unspecified_argument:
			default = self._unknown
		fget, default = self._check_fget(fget, default)
		super().__init__(**kwargs)
		self.attrname = attrname
		# if not cache:
		# 	raise NotImplementedError('cache=False not implemented') # TODO: implement
		self.cache = cache
		self.default = default
		self.fget = fget
		self.fformat = fformat
		self.fset = fset
		self.fdel = fdel


	def _check_fget(self, fget=None, default=unspecified_argument):
		if fget is None:
			if callable(default) and not isinstance(default, type) and default.__qualname__ != default.__name__:
				return default, self._unknown  # decorator has no other args
			return None, default  # no fget provided (optionally can be added with __call__)
		return fget, default  # fget was specified as keyword argument


	def setup(self, owner: Type, name: str) -> 'method_decorator':
		if self.attrname is None:
			self.attrname = name
		elif name != self.attrname:
			raise TypeError(
				"Cannot assign the same cached_property to two different names "
				f"({self.attrname!r} and {name!r})."
			)
		return self


	def __set_name__(self, owner, name):
		self.setup(owner, name)


	_MissingValueError = MissingValueError


	def __get__(self, instance, owner=None):
		if instance is None:
			return self
		return self.get_value(instance, owner=owner)


	def is_missing(self, instance):
		return self.attrname not in instance.__dict__ and self.default is self._unknown and self.fget is None


	def is_cached(self, instance):
		return self.attrname in instance.__dict__


	def get_value(self, instance=None, owner=None):
		if instance is None:
			if self.default is self._unknown:
				raise self._MissingValueError(f'No default value for {self.attrname!r} (must include instance)')
			return self.default
		if self.attrname is None:
			raise TypeError(
				"Cannot use cached_property instance without calling __set_name__ on it.")
		try:
			cache = instance.__dict__
		except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
			msg = (
				f"No '__dict__' attribute on {type(instance).__name__!r} "
				f"instance to cache {self.attrname!r} property."
			)
			raise TypeError(msg) from None
		val = cache.get(self.attrname, self._unknown)
		if val is self._unknown: # TODO: make thread safe
			if self.fget is None:
				if self.default is self._unknown:
					raise self._MissingValueError(f'No value for {self.attrname!r} property')
				return self.default
			val = self.fget.__get__(instance, owner)()
			if self.cache:
				try:
					cache[self.attrname] = val
				except TypeError:
					msg = (
						f"The '__dict__' attribute on {type(instance).__name__!r} instance "
						f"does not support item assignment for caching {self.attrname!r} property."
					)
					raise TypeError(msg) from None
		return val


	def __set__(self, instance, value):
		self.update_value(instance, value)


	_IgnoreUpdateFlag = IgnoreUpdateFlag

	def update_value(self, instance, value):
		if self.fformat is not None:
			try:
				value = self.fformat.__get__(instance, type(instance))(value)
			except self._IgnoreUpdateFlag:
				return

		if self.fset is None:
			return instance.__dict__.setdefault(self.attrname, value)

		try:
			setter = self.fset.__get__(instance, type(instance))
		except AttributeError:
			setter = self.fset

		setter(value)


	def __delete__(self, instance):
		self.remove_value(instance)


	def remove_value(self, instance):
		if self.fdel is None:
			return delattr(instance, self.attrname)

		try:
			deleter = self.fdel.__get__(instance, type(instance))
		except AttributeError:
			deleter = self.fdel

		deleter()


	def __call__(self, fn):
		return self.getter(fn)


	def copy(self, **kwargs):
		return self.__class__(default=self.default,
		                      fget=self.fget, fset=self.fset, fdel=self.fdel, fformat=self.fformat,
		                      cache=self.cache, attrname=self.attrname, **kwargs)


	def getter(self, fget):
		new = self.copy()
		new.fget = fget
		return new


	def formatter(self, fformat):
		new = self.copy()
		new.fformat = fformat
		return new


	def setter(self, fset):
		new = self.copy()
		new.fset = fset
		return new


	def deleter(self, fdel):
		new = self.copy()
		new.fdel = fdel
		return new



# TODO: add 'help' option to include custom info about the hyperparameter

class HyperparameterBase(DefaultProperty, AbstractHyperparameter): # TODO: add 'cached' option
	def __init__(self, default=unspecified_argument, *, space=None, hidden=None, required=None, **kwargs):
		super().__init__(default=default, **kwargs)
		self.space = space
		self.hidden = hidden
		self.required = required


	def setup(self, owner, name):
		out = super().setup(owner, name)
		if self.hidden is None and self.attrname.startswith('_'):
			self.hidden = True
		return out



class InheritableHyperparameter(HyperparameterBase):
	def __init__(self, default=unspecified_argument, *, inherit=None, **kwargs):
		super().__init__(default=default, **kwargs)
		self.inherit = inherit




















import sys
from omnibelt import unspecified_argument


class SharableAttrs:
	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self._shared_attrs = set()


	def _make_shared(self, key, val=unspecified_argument):
		if val is unspecified_argument:
			val = getattr(self, key)
		self._shared_attrs.add(key)
		setattr(self, key, [val])


	def access_shared(self, key):
		val = getattr(self, key, None)
		if val is not None and key in self._shared_attrs:
			return val[0]
		return val



if 'torch' in sys.modules:
	import torch
	_default_device = 'cuda' if torch.cuda.is_available() else 'cpu'
else:
	_default_device = 'cpu'



class Device:
	def __init__(self, *args, device=_default_device, **kwargs):
		super().__init__(*args, **kwargs)
		self.device = device


	def to(self, device, **kwargs):
		self.device = device
		out = self._to(device, **kwargs)
		if out is not None:
			return out
		return self


	def _to(self, device, **kwargs):
		raise NotImplementedError



class DeviceContainer(Device):
	def __init__(self, *, children=None, **kwargs):
		if children is None:
			children = {}
		super().__init__(**kwargs)
		self._device_children = set()
		self._register_deviced_children(**children)


	def _register_deviced_children(self, **children):
		for name, child in children.items():
			self._device_children.add(name)
			setattr(self, name, child)


	def _to(self, device, **kwargs):
		children = {}
		for name in self._device_children:
			obj = getattr(self, name, None)
			if obj is not None:
				children[name] = obj.to(device)
		self._register_deviced_children(**children)










from typing import Any, Optional, Iterator
from collections import UserDict
from omnibelt import filter_duplicates

from .abstract import AbstractGang, AbstractGig
from .errors import GadgetFailure, ApplicationAmbiguityError
from .gaggles import GaggleBase
from .gigs import GigBase, GroupCache


class GroupBase(GaggleBase, AbstractGang):
	"""
	The GroupBase class is a subclass of GaggleBase and AbstractGroup. It provides methods to handle gizmo grabbing and packaging.

	Attributes:
		_current_context (Optional[AbstractGig]): The current context of the group.
	"""

	_current_context: Optional[AbstractGig]

	def __init__(self, *args, gap: Optional[dict[str, str]] = None, **kwargs):
		"""
		Initializes a new instance of the GroupBase class.

		Args:
			args: Variable length argument list.
			gap (Optional[dict[str, str]]): A dictionary of gizmo mappings. If not provided, an empty dictionary will be used.
			kwargs: Arbitrary keyword arguments.
		"""
		if gap is None:
			gap = {}
		super().__init__(*args, **kwargs)
		self._raw_gap = gap # internal gizmos -> external gizmos
		self._raw_reverse_gap = None
		self._gig_stack = []

	def _gizmos(self) -> Iterator[str]:
		"""
		Lists gizmos produced by self using internal names.

		Returns:
			Iterator[str]: An iterator over the gizmos.
		"""
		yield from super().gizmos()

	def gizmos(self) -> Iterator[str]:
		"""
		Lists gizmos produced by self using external names.

		Returns:
			Iterator[str]: An iterator over the gizmos.
		"""
		for gizmo in self._gizmos():
			yield self.gizmo_to(gizmo)

	@property
	def internal2external(self) -> dict[str, str]:
		"""
		Getter for the internal to external gizmo mapping.

		Returns:
			dict[str, str]: The internal to external gizmo mapping.
		"""
		return self._raw_gap

	@internal2external.setter
	def internal2external(self, value: dict[str, str]):
		"""
		Setter for the internal to external gizmo mapping.

		Args:
			value (dict[str, str]): The new internal to external gizmo mapping.
		"""
		self._raw_gap = value
		self._raw_reverse_gap = None

	@property
	def external2internal(self) -> dict[str, str]:
		"""
		Getter for the external to internal gizmo mapping.

		Returns:
			dict[str, str]: The external to internal gizmo mapping.
		"""
		if self._raw_reverse_gap is None:
			self._raw_reverse_gap = self._infer_external2internal(self._raw_gap, self._gizmos())
		return self._raw_reverse_gap

	@staticmethod
	def _infer_external2internal(raw: dict[str, str], products: Iterator[str]) -> dict[str, str]:
		"""
		Infers the external to internal gizmo mapping from the provided raw mapping and products.

		Args:
			raw (dict[str, str]): The raw gizmo mapping.
			products (Iterator[str]): An iterator over the products.

		Returns:
			dict[str, str]: The inferred external to internal gizmo mapping.
		"""
		reverse = {}
		for product in products:
			if product in raw:
				external = raw[product]
				if external in reverse:
					raise ApplicationAmbiguityError(product, [reverse[external], product])
				reverse[external] = product
		return reverse

	def gizmo_from(self, gizmo: str) -> str:
		"""
		Converts an external gizmo to its internal representation.

		Args:
			gizmo (str): The external gizmo.

		Returns:
			str: The internal representation of the gizmo.
		"""
		return self.external2internal.get(gizmo, gizmo)

	def gizmo_to(self, gizmo: str) -> str:
		"""
		Converts an internal gizmo to its external representation.

		Args:
			gizmo (str): The internal gizmo.

		Returns:
			str: The external representation of the gizmo.
		"""
		return self.internal2external.get(gizmo, gizmo)

	def _grab(self, gizmo: str):
		"""
		Grabs a gizmo from self.

		Args:
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The grabbed gizmo.
		"""
		return super().grab_from(self, gizmo)

	def grab_from(self, ctx: AbstractGig, gizmo: str) -> Any:
		"""
		Tries to grab a gizmo from the context.

		Args:
			ctx (Optional[AbstractGig]): The context from which to grab the gizmo.
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The grabbed gizmo.
		"""
		if ctx is not None and ctx is not self:
			self._gig_stack.append(ctx)
			gizmo = self.gizmo_from(gizmo)  # convert to internal gizmo

		try:
			out = self._grab(gizmo)
		except self._GadgetFailure:
			if len(self._gig_stack) == 0 or ctx is self._gig_stack[-1]:
				raise
			# default to parent/s
			out = self._gig_stack[-1].grab(self.gizmo_to(gizmo))

		if len(self._gig_stack) and ctx is self._gig_stack[-1]:
			self._gig_stack.pop()

		return out

	# def _grab_from_fallback(self, error: Exception, ctx: Optional[AbstractGig], gizmo: str) -> Any:
	# 	assert ctx is self, f'{ctx} != {self}'
	# 	if len(self._gig_stack):
	# 		return super()._grab_from_fallback(error, self._gig_stack[-1], self.gizmo_to(gizmo))
	# 	raise error from error
	#
	#
	# def grab_from(self, ctx: Optional[AbstractGig], gizmo: str) -> Any:
	# 	if ctx is not None and ctx is not self:
	# 		self._gig_stack.append(ctx)
	# 		gizmo = self.gizmo_from(gizmo) # convert to internal gizmo
	# 	out = super().grab_from(self, gizmo)
	# 	if len(self._gig_stack) and ctx is self._gig_stack[-1]:
	# 		self._gig_stack.pop()
	# 	return out


class CachableGroup(GroupBase):
	"""
	The CachableGroup class is a subclass of GroupBase. It provides methods to handle gizmo grabbing with cache support.

	Attributes:
		_GroupCacheMiss (KeyError): The exception to be raised when a cache miss occurs.
	"""

	_GroupCacheMiss = KeyError

	def _grab(self, gizmo: str) -> Any:
		"""
		Tries to grab a gizmo from the group. If the gizmo is not found in the group's cache, it checks the cache using
		the external gizmo name. If it still can't be found in any cache, it grabs it from the group's gadgets.

		Args:
			gizmo (str): The name of the gizmo to grab.

		Returns:
			Any: The grabbed gizmo.
		"""
		if len(self._gig_stack):
			# check cache (if one exists)
			for parent in reversed(self._gig_stack):
				if isinstance(parent, GroupCache):
					try:
						return parent.check_group_cache(self, gizmo)
					except self._GroupCacheMiss:
						pass

			# if it can't be found in my cache, check the cache using the external gizmo name
			ext = self.gizmo_to(gizmo)
			for parent in reversed(self._gig_stack):
				if isinstance(parent, GroupCache) and parent.is_cached(ext):
					return parent.grab(ext)

		# if it can't be found in any cache, grab it from my gadgets
		out = super()._grab(gizmo)

		# update my cache
		if len(self._gig_stack):
			for parent in reversed(self._gig_stack):
				if isinstance(parent, GroupCache):
					parent.update_group_cache(self, gizmo, out)
					break

		return out


class SelectiveGroup(GroupBase):
	"""
	The SelectiveGroup class is a subclass of GroupBase. It provides methods to handle selective gizmo mapping.

	Args:
		args: Variable length argument list.
		gap (dict[str, str] | list[str] | None): A dictionary or list of gizmo mappings. If not provided, an empty dictionary will be used.
		kwargs: Arbitrary keyword arguments.
	"""

	def __init__(self, *args, gap: dict[str, str] | list[str] | None = None, **kwargs):
		"""
		Initializes a new instance of the SelectiveGroup class.

		If the gap argument is a list, it is converted to a dictionary with the same keys and values.
		If the gap argument is a dictionary, it is processed to ensure that all values are not None.

		Args:
			args: Variable length argument list.
			gap (dict[str, str] | list[str] | None): A dictionary or list of gizmo mappings. If not provided, an empty dictionary will be used.
			kwargs: Arbitrary keyword arguments.
		"""
		if isinstance(gap, list):
			gap = {gizmo: gizmo for gizmo in gap}
		if isinstance(gap, dict):
			gap = {gizmo: gizmo if ext is None else ext for gizmo, ext in gap.items()}
		super().__init__(*args, gap=gap, **kwargs)

	def gizmos(self) -> Iterator[str]:
		"""
		Lists gizmos produced by self using external names.

		Returns:
			Iterator[str]: An iterator over the gizmos.
		"""
		for gizmo in self._gizmos():
			if gizmo in self.internal2external:
				yield self.gizmo_to(gizmo)





# from .old_abstract import AbstractData, AbstractView, AbstractBuffer
# from .buffers import Buffer, RemoteBuffer, HDFBuffer, BufferView, ReplacementBuffer, TransformedBuffer, NarrowBuffer
# from .base import DataCollection, DataSource, Dataset, Batch
# from .flavors import SimpleDataset, GenerativeDataset, ObservationDataset, \
# 	SupervisedDataset, LabeledDataset, SyntheticDataset, \
# 	RootedDataset, EncodableDataset, DownloadableDataset, ImageDataset

from .top import Datastream, Dataset, Buffer, SimpleDataset, BudgetLoader
# from .materials import material
from . import toy
from . import flavors


# TODO: move to foundation

# from omnilearn.datasets.disentangling import dSprites, Shapes3D, MPI3D, CelebA


from omnibelt import unspecified_argument

from ..persistent import Fingerprinted



class Function(Fingerprinted):
	din, dout = None, None

	def __init__(self, *args, din=unspecified_argument, dout=unspecified_argument, **kwargs):
		super().__init__(*args, **kwargs)
		if din is not unspecified_argument:
			self.din = din
		if dout is not unspecified_argument:
			self.dout = dout
	# self.din, self.dout = din, dout


	def get_dims(self):
		return self.din, self.dout


	def __call__(self, inp):
		return self.forward(inp)


	# def forward(self, inp):
	# 	raise NotImplementedError


	def _fingerprint_data(self):
		data = super()._fingerprint_data()
		data['din'] = self.din
		data['dout'] = self.dout
		# if self.din is not None:
		# 	x = self.din.sample(4, gen=torch.Generator().manual_seed(16283393149723337453))
		# 	try:
		# 		with torch.no_grad():
		# 			y = self(x)
		# 		data['output'] = extractor.extract_data(y)
		# 	except:
		# 		raise # TESTING
		return data
















import logging
import yaml
from typing import Optional
from collections import OrderedDict
from .abstract import AbstractGadgetError, AbstractGadget

logger = logging.getLogger('omniply')



class GadgetFailure(AbstractGadgetError):
	'''General error for when a gadget fails to grab a gizmo'''
	def __init__(self, message: Optional[str] = None):
		super().__init__(message)
		self.message = message

	def __hash__(self):
		return hash(repr(self))


	def __eq__(self, other):
		return repr(self) == repr(other)


	@property
	def description(self) -> str:
		return str(self)



class MissingGadget(GadgetFailure, KeyError):
	'''Error for when a gadget fails to grab a gizmo because the gadget can't find it'''
	def __init__(self, gizmo: str, *, message: Optional[str] = None):
		if message is None:
			message = gizmo
		super().__init__(message)
		self.gizmo = gizmo

	@property
	def description(self) -> str:
		return f'missing gadget for {self.gizmo!r}'



class AssemblyError(GadgetFailure):
	'''Error for when a gadget fails to grab a gizmo because the gizmo can't be assembled from the gadgets available'''
	def __init__(self, failures: OrderedDict[GadgetFailure, AbstractGadget], *,
				 message: Optional[str] = None):
		if message is None:
			errors = [str(error) for error in failures]
			message = f'{len(errors)} failures: {", ".join(errors)}'
		super().__init__(message)
		self.failures = failures



class GrabError(AbstractGadgetError):
	def __init__(self, gizmo: str, error: AbstractGadgetError, *, message: Optional[str] = None):
		if message is None:
			message = f'{gizmo!r} failed due to: {error.description}'
		super().__init__(message)
		self.error = error
		self.gizmo = gizmo

	@property
	def description(self) -> str:
		return str(self)



class ApplicationAmbiguityError(ValueError):
	def __init__(self, gizmo: str, options: list[str], *, message: Optional[str] = None):
		if message is None:
			message = f'{gizmo!r} cant choose between: {options}'
		super().__init__(message)
		self.gizmo = gizmo
		self.options = options






from .tools import *



# class KitBase(AbstractToolKit, ToolBase):
# 	pass



class Kit(AbstractToolKit, ToolBase):
	_tools_table: Dict[str, List[AbstractGadget]] # tools are kept in O-N order (reversed) for easy updates

	def __init__(self, *args, tools_table: Optional[Mapping] = None, **kwargs):
		if tools_table is None:
			tools_table = {}
		super().__init__(*args, **kwargs)
		self._tools_table = tools_table


	def gizmos(self) -> Iterator[str]:
		yield from self._tools_table.keys()


	def gives(self, gizmo: str) -> bool:
		return gizmo in self._tools_table


	def _vendors(self, gizmo: Optional[str] = None) -> Iterator[AbstractGadget]:
		if gizmo is None:
			for tool in filter_duplicates(chain.from_iterable(map(reversed, self._tools_table.values()))):
				yield from tool.vendors(gizmo)
		else:
			if gizmo not in self._tools_table:
				raise self._MissingGizmoError(gizmo)
			for tool in filter_duplicates(reversed(self._tools_table[gizmo])):
				yield from tool.vendors(gizmo)


	_AssemblyFailedError = AssemblyFailedError
	def grab_from(self, ctx: 'AbstractGig', gizmo: str) -> Any:
		failures = []
		for tool in self._vendors(gizmo):
			try:
				return tool.grab_from(ctx, gizmo)
			except GadgetFailedError as e:
				e.tool = tool
				failures.append(e)
			except:
				prt.debug(f'{tool!r} failed while trying to produce: {gizmo!r}')
				raise
		if failures:
			raise self._AssemblyFailedError(gizmo, *failures)
		raise self._ToolFailedError(gizmo)



class LoopyKit(Kit):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._grabber_stack: Optional[Dict[str, Iterator[AbstractGadget]]] = {}


	def grab_from(self, ctx: 'AbstractGig', gizmo: str) -> Any:
		failures = []
		itr = self._grabber_stack.setdefault(gizmo, self._vendors(gizmo))
		# should be the same as Kit, except the iterators are cached until the gizmo is produced
		for tool in itr:
			try:
				out = tool.grab_from(ctx, gizmo)
			except GadgetFailedError as e:
				e.tool = tool
				failures.append(e)
			except:
				prt.debug(f'{tool!r} failed while trying to produce: {gizmo!r}')
				raise
			else:
				if gizmo in self._grabber_stack:
					self._grabber_stack.pop(gizmo)
				return out
		if gizmo in self._grabber_stack:
			self._grabber_stack.pop(gizmo)
		if failures:
			raise self._AssemblyFailedError(gizmo, *failures)
		raise self._ToolFailedError(gizmo)



class MutableKit(Kit):
	def include(self, *tools: Union[AbstractGadget, Callable]) -> 'MutableKit': # TODO: return Self
		'''adds given tools in reverse order'''
		return self.extend(tools)
		
		
	def extend(self, tools: Iterable[AbstractGadget]) -> 'MutableKit':
		new = {}
		for tool in tools:
			for gizmo in tool.gizmos():
				new.setdefault(gizmo, []).append(tool)
		for gizmo, tools in new.items():
			if gizmo in self._tools_table:
				for tool in tools:
					if tool in self._tools_table[gizmo]:
						self._tools_table[gizmo].remove(tool)
			self._tools_table.setdefault(gizmo, []).extend(reversed(tools))
		return self


	def exclude(self, *tools: AbstractGadget) -> 'MutableKit':
		'''removes the given tools, if they are found'''
		for tool in tools:
			for gizmo in tool.gizmos():
				if gizmo in self._tools_table and tool in self._tools_table[gizmo]:
					self._tools_table[gizmo].remove(tool)
		return self
	


class CraftyKit(Kit, InheritableCrafty):
	def _process_crafts(self):
		# avoid duplicate keys (if you overwrite a method, only the last one will be used)
		items = OrderedDict()
		for src, key, craft in self._emit_all_craft_items(): # N-O
			if key not in items:
				items[key] = craft
			
		# convert crafts to skills and add in O-N order
		table = {}
		for key, craft in reversed(items.items()): # N-O
			skill = craft.as_skill(self)
			for gizmo in skill.gizmos():
				table.setdefault(gizmo, []).append(skill)
			
		# add N-O skills in reverse order for O-N _tools_table
		for gizmo, tools in table.items(): # tools is N-O
			self._tools_table.setdefault(gizmo, []).extend(tools) # added in O-N order





import os
from pathlib import Path
from omnibelt import unspecified_argument, agnostic, agnosticproperty


class Rooted:
	_DEFAULT_MASTER_ROOT = os.getenv('OMNIDATA_PATH', 'local_data/')

	@agnostic
	def _infer_root(self, root=None):
		if root is None:
			root = self._DEFAULT_MASTER_ROOT
		root = Path(root)
		# os.makedirs(str(root), exist_ok=True)
		return root


	@agnosticproperty
	def root(self):
		return self._infer_root()
















import numpy as np
# import torch
from omnibelt import unspecified_argument, agnostic
# from sklearn.datasets import make_swiss_roll

# from ..materials import Materialed, material
from ...structure import spaces, Decoder, Generator, NormalDistribution
from ...features import Seeded
from ...parameters import submodule, hparam, inherit_hparams
from ...tools import material, machine, space

from ..flavors import Synthetic, Sampledstream, SimpleSampledStream
from ..top import Datastream



class ManifoldStream(Datastream, Seeded, Generator):
	def generate(self, N: int): # generates observations
		return self.generate_observation_from_manifold(self.sample_manifold(N))


	@material.from_size('manifold')
	def sample_manifold(self, N: int):
		return self.space_of('manifold').sample(N)


	@property
	def manifold_space(self):
		return self.space_of('manifold')


	@machine('observation')
	def generate_observation_from_manifold(self, manifold):
		raise NotImplementedError



class DeterministicManifold(ManifoldStream, Decoder):
	def generate_observation_from_manifold(self, manifold):
		return self.decode(manifold)


	def decode(self, manifold): # deterministic mapping from manifold to observation
		raise NotImplementedError



class DistributionalManifold(ManifoldStream):
	def generate_observation_from_manifold(self, manifold):
		with self.force_rng(rng=self.rng): # TODO: change force to push
			return self.decode_distribution(manifold).sample()


	def decode_distribution(self, manifold):
		raise NotImplementedError



class NoisyManifold(DistributionalManifold, DeterministicManifold):
	def _decode_distrib_from_mean(self, mean):
		raise NotImplementedError


	def decode_distribution(self, manifold):
		return self._decode_distrib_from_mean(self.decode(manifold))



class Noisy(NoisyManifold):
	noise_std = hparam(0.1, space=spaces.HalfBound(min=0.))


	def _decode_distrib_from_mean(self, mean):
		return NormalDistribution(mean, self.noise_std * torch.ones_like(mean))



####################################################################################################



class SwissRoll(DeterministicManifold):
	Ax = hparam(np.pi / 2, space=spaces.HalfBound(min=0.))
	Ay = hparam(21., space=spaces.HalfBound(min=0.))
	Az = hparam(np.pi / 2, space=spaces.HalfBound(min=0.))

	freq = hparam(0.5, space=spaces.HalfBound(min=0.))
	tmin = hparam(3., space=spaces.HalfBound(min=0.))
	tmax = hparam(9., space=spaces.HalfBound(min=0.))


	@space('manifold')
	def manifold_space(self):
		return spaces.Joint(
			spaces.Bound(min=self.tmin, max=self.tmax),
			spaces.Bound(min=0., max=1.),
		)


	@space('target')
	def target_space(self):
		return self.manifold_space[0]


	@space('observation')
	def observation_space(self):
		# assert Ax > 0 and Ay > 0 and Az > 0 and freq > 0 and tmax > tmin, \
		# 	f'invalid parameters: {Ax} {Ay} {Az} {freq} {tmax} {tmin}'
		return spaces.Joint(
			spaces.Bound(min=-self.Ax * self.tmax, max=self.Ax * self.tmax),
			spaces.Bound(min=0., max=self.Ay),
			spaces.Bound(min=-self.Az * self.tmax, max=self.Az * self.tmax),
		)


	@machine('target')
	def get_target(self, manifold):
		return manifold.narrow(-1,0,1) #if self.target_theta else manifold


	def decode(self, manifold):
		theta = manifold.narrow(-1,0,1)
		height = manifold.narrow(-1,1,1)

		pts = torch.cat([
			self.Ax * theta * theta.mul(self.freq*np.pi).cos(),
			self.Ay * height,
			self.Az * theta * theta.mul(self.freq*np.pi).sin(),
		], -1)
		return pts



class Helix(DeterministicManifold):
	n_helix = hparam(2, space=spaces.Naturals())

	periodic_strand = hparam(False, space=spaces.Binary())

	Rx = hparam(1., space=spaces.HalfBound(min=0.))
	Ry = hparam(1., space=spaces.HalfBound(min=0.))
	Rz = hparam(1., space=spaces.HalfBound(min=0.))

	w = hparam(1., space=spaces.HalfBound(min=0.))


	@space('manifold')
	def manifold_space(self):
		return spaces.Joint(
			spaces.Periodic(min=-1., max=1.) if self.periodic_strand else spaces.Bound(min=-1., max=1.),
			spaces.Categorical(n=self.n_helix),
		)


	@space('target')
	def target_space(self):
		return self.manifold_space[-1]


	@space('observation')
	def observation_space(self):
		return spaces.Joint(
			spaces.Bound(min=-self.Rx, max=self.Rx),
			spaces.Bound(min=-self.Ry, max=self.Ry),
			spaces.Bound(min=-self.Rz, max=self.Rz),
		)


	@machine('target')
	def get_target(self, manifold):
		return manifold.narrow(-1,1,1).long()


	def decode(self, manifold):
		z = manifold.narrow(-1, 0, 1)
		n = manifold.narrow(-1, 1, 1)
		theta = z.mul(self.w).add(n.div(self.n_helix) * 2).mul(np.pi)

		amp = torch.as_tensor([self.Rx, self.Ry, self.Rz]).float().to(n.device)
		pts = amp.unsqueeze(0) * torch.cat([theta.cos(), theta.sin(), z], -1)
		return pts



@inherit_hparams('n_samples')
class SwissRollDataset(SimpleSampledStream, source_cls=SwissRoll):
	pass



@inherit_hparams('n_samples')
class HelixDataset(SimpleSampledStream, source_cls=Helix):
	pass


